import subprocess
from pathlib import Path
from typing import List, Optional

import typer
from google.api_core import exceptions
from google.cloud.notebooks_v1.services.notebook_service import NotebookServiceClient
from google.cloud.notebooks_v1.types import ContainerImage, CreateInstanceRequest, Instance, VmImage
from waiting import wait

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.notebook import NotebookModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService
from wanna.core.services.docker import DockerService
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils import templates
from wanna.core.utils.gcp import construct_vm_image_family_from_vm_image, upload_string_to_gcs

logger = get_logger(__name__)


class NotebookService(BaseService[NotebookModel]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        owner: Optional[str] = None,
        version: str = "dev",
    ):
        super().__init__(
            instance_type="notebook",
        )
        self.version = version
        self.instances = config.notebooks
        self.wanna_project = config.wanna_project
        self.bucket_name = config.gcp_profile.bucket
        self.notebook_client = NotebookServiceClient()
        self.config = config
        self.docker_service = (
            DockerService(
                docker_model=config.docker,
                gcp_profile=config.gcp_profile,
                version=version,
                work_dir=workdir,
                wanna_project_name=self.wanna_project.name,
            )
            if config.docker
            else None
        )

        self.owner = owner
        self.tensorboard_service = TensorboardService(config=config)

    def _delete_one_instance(self, notebook_instance: NotebookModel) -> None:
        """
        Delete one notebook instance. This assumes that it has been already verified that notebook exists.

        Args:
            notebook_instance: notebook to delete
        """

        exists = self._instance_exists(notebook_instance)
        if exists:
            with logger.user_spinner(f"Deleting {self.instance_type} {notebook_instance.name}"):
                deleted = self.notebook_client.delete_instance(
                    name=f"projects/{notebook_instance.project_id}/locations/"
                    f"{notebook_instance.zone}/instances/{notebook_instance.name}"
                )
                deleted.result()
        else:
            logger.user_error(
                f"Notebook with name {notebook_instance.name} was not found in region {notebook_instance.region}",
            )
            typer.Exit(1)

    def _create_one_instance(self, instance: NotebookModel, **kwargs) -> None:
        """
        Create a notebook instance based on information in NotebookModel class.
        1. Check if the notebook already exists
        2. Parse the information from NotebookModel to GCP API friendly format = instance_request
        3. Wait for the compute instance behind the notebook to start
        4. Wait for JupyterLab to start
        5. Get and print the link to JupyterLab

        Args:
            instance: notebook to be created

        """
        exists = self._instance_exists(instance)
        if exists:
            logger.user_info(f"Instance {instance.name} already exists in location {instance.zone}")
            should_recreate = typer.confirm("Are you sure you want to delete it and start a new?")
            if should_recreate:
                self._delete_one_instance(instance)
            else:
                return
        instance_request = self._create_instance_request(notebook_instance=instance, deploy=True)
        with logger.user_spinner(f"Creating underlying compute engine instance for {instance.name}"):
            nb_instance = self.notebook_client.create_instance(instance_request)
            instance_full_name = (
                nb_instance.result().name
            )  # .result() waits for compute engine behind the notebook to start
        with logger.user_spinner("Starting JupyterLab"):
            wait(
                lambda: self._validate_jupyterlab_state(instance_full_name, Instance.State.ACTIVE),
                timeout_seconds=450,
                sleep_seconds=20,
                waiting_for="Starting JupyterLab in your instance",
            )
            jupyterlab_link = self._get_jupyterlab_link(instance_full_name)
        logger.user_success(f"JupyterLab started at {jupyterlab_link}")

    def _list_running_instances(self, project_id: str, location: str) -> List[str]:
        """
        List all notebooks with given project_id and location.

        Args:
            project_id: GCP project ID
            location: GCP location (zone)

        Returns:
            instance_names: List of the full names on notebook instances (this includes project_id, and zone)

        """
        instances = self.notebook_client.list_instances(parent=f"projects/{project_id}/locations/{location}")
        instance_names = [i.name for i in instances.instances]
        return instance_names

    def _instance_exists(self, instance: NotebookModel) -> bool:
        """
        Check if the instance with given instance_name exists in given GCP project project_id and location.
        Args:
            instance: notebook to verify if exists on GCP

        Returns:
            True if exists, False if not
        """
        full_instance_name = f"projects/{instance.project_id}/locations/{instance.zone}/instances/{instance.name}"
        return full_instance_name in self._list_running_instances(instance.project_id, instance.zone)

    def _create_instance_request(self, notebook_instance: NotebookModel, deploy: bool = True) -> CreateInstanceRequest:
        """
        Transform the information about desired notebook from our NotebookModel model (based on yaml config)
        to the form suitable for GCP API.

        Args:
            notebook_instance

        Returns:
            CreateInstanceRequest
        """

        # Network
        full_network_name = self._get_resource_network(
            project_id=self.config.gcp_profile.project_id,
            push_mode=PushMode.all,
            resource_network=notebook_instance.network,
            fallback_project_network=self.config.gcp_profile.network,
            use_project_number=True,
        )
        subnet = notebook_instance.subnet if notebook_instance.subnet else self.config.gcp_profile.subnet
        full_subnet_name = self._get_resource_subnet(
            full_network_name,
            subnet,
            notebook_instance.region,
        )

        # GPU
        if notebook_instance.gpu:
            accelerator_config = Instance.AcceleratorConfig(
                core_count=notebook_instance.gpu.count,
                type_=notebook_instance.gpu.accelerator_type,
            )
            install_gpu_driver = notebook_instance.gpu.install_gpu_driver
        else:
            accelerator_config = None
            install_gpu_driver = False
        # Environment
        if notebook_instance.environment.docker_image_ref:
            if self.docker_service:
                vm_image = None
                image_tag = self.docker_service.get_image(
                    docker_image_ref=notebook_instance.environment.docker_image_ref
                )
                if image_tag[1]:
                    self.docker_service.push_image(image_tag[1])
                repository = image_tag[2].partition(":")[0]
                tag = image_tag[2].partition(":")[-1]
                container_image = ContainerImage(
                    repository=repository,
                    tag=tag,
                )
            else:
                raise Exception("Docker params in wanna-ml config not defined")
        elif notebook_instance.environment.vm_image:
            vm_image = VmImage(
                project="deeplearning-platform-release",
                image_family=construct_vm_image_family_from_vm_image(
                    notebook_instance.environment.vm_image.framework,
                    notebook_instance.environment.vm_image.version,
                    notebook_instance.environment.vm_image.os,
                ),
            )
            container_image = None
        else:
            raise ValueError(
                "No notebook environment was found. This should not be possible."
                " Something went wrong during model validation"
            )
        # Disks
        boot_disk_type = notebook_instance.boot_disk.disk_type if notebook_instance.boot_disk else None
        boot_disk_size_gb = notebook_instance.boot_disk.size_gb if notebook_instance.boot_disk else None
        data_disk_type = notebook_instance.data_disk.disk_type if notebook_instance.data_disk else None
        data_disk_size_gb = notebook_instance.data_disk.size_gb if notebook_instance.data_disk else None
        disk_encryption = "CMEK" if self.config.gcp_profile.kms_key else None
        kms_key = self.config.gcp_profile.kms_key if self.config.gcp_profile.kms_key else None

        # service account and instance owners
        service_account = notebook_instance.service_account
        instance_owner = self.owner or notebook_instance.instance_owner
        instance_owners = [instance_owner] if instance_owner else None

        # labels and tags
        tags = notebook_instance.tags
        labels = {
            "wanna_name": notebook_instance.name,
            "wanna_resource": self.instance_type,
        }
        if notebook_instance.labels:
            labels = {**notebook_instance.labels, **labels}

        # post startup script
        if deploy and (notebook_instance.bucket_mounts or notebook_instance.tensorboard_ref):
            script = self._prepare_startup_script(self.instances[0])
            blob = upload_string_to_gcs(
                script,
                notebook_instance.bucket or self.bucket_name,
                f"notebooks/{notebook_instance.name}/startup_script.sh",
            )
            post_startup_script = f"gs://{blob.bucket.name}/{blob.name}"
        else:
            post_startup_script = None

        instance = Instance(
            vm_image=vm_image,
            container_image=container_image,
            machine_type=notebook_instance.machine_type,
            accelerator_config=accelerator_config,
            install_gpu_driver=install_gpu_driver,
            network=full_network_name,
            subnet=full_subnet_name,
            boot_disk_type=boot_disk_type,
            boot_disk_size_gb=boot_disk_size_gb,
            data_disk_type=data_disk_type,
            data_disk_size_gb=data_disk_size_gb,
            post_startup_script=post_startup_script,
            instance_owners=instance_owners,
            service_account=service_account,
            tags=tags,
            labels=labels,
            disk_encryption=disk_encryption,
            kms_key=kms_key,
            no_public_ip=notebook_instance.no_public_ip,
            no_proxy_access=notebook_instance.no_proxy_access,
        )

        return CreateInstanceRequest(
            parent=f"projects/{notebook_instance.project_id}/locations/{notebook_instance.zone}",
            instance_id=notebook_instance.name,
            instance=instance,
        )

    def _prepare_startup_script(self, nb_instance: NotebookModel) -> str:
        """
        Prepare the notebook startup script based on the information from notebook.
        This script run at the Compute Engine Instance creation time with a root user.

        Args:
            nb_instance

        Returns:
            startup_script
        """
        bucket_mounts = nb_instance.bucket_mounts
        if nb_instance.tensorboard_ref:
            tensorboard_resource_name = self.tensorboard_service.get_or_create_tensorboard_instance_by_name(
                nb_instance.tensorboard_ref
            )
        else:
            tensorboard_resource_name = None
        startup_script = templates.render_template(
            Path("notebook_startup_script.sh.j2"),
            bucket_mounts=bucket_mounts,
            tensorboard_resource_name=tensorboard_resource_name,
        )
        return startup_script

    def _validate_jupyterlab_state(self, instance_id: str, state: int) -> bool:
        """
        Validate if the given notebook instance is in given state.

        Args:
            instance_id: Full notebook instance id
            state: Notebook state (ACTIVE, PENDING,...)

        Returns:
            True if desired state, False otherwise
        """
        try:
            instance_info = self.notebook_client.get_instance(name=instance_id)
        except exceptions.NotFound:
            raise exceptions.NotFound(f"Notebook {instance_id} was not found.") from None
        return instance_info.state == state

    def _get_jupyterlab_link(self, instance_id: str) -> str:
        """
        Get a link to jupyterlab proxy based on given notebook instance id.
        Args:
            instance_id: full notebook instance id

        Returns:
            proxy_uri: link to jupyterlab
        """
        instance_info = self.notebook_client.get_instance({"name": instance_id})
        return f"https://{instance_info.proxy_uri}"

    def _ssh(self, notebook_instance: NotebookModel, run_in_background: bool, local_port: int) -> None:
        """
        SSH connect to the notebook instance if the instance is already started.

        Args:
            notebook_instance: notebook model representing the instance you want to connect to
            run_in_background: whether to run in the background or in interactive mode
            local_port: jupyter lab will be exposed at this port at localhost

        Returns:

        """
        exists = self._instance_exists(notebook_instance)
        if not exists:
            logger.user_info(
                f"Notebook {notebook_instance.name} is not running, create it first and then ssh connect to it."
            )
            return

        bash_command = f"gcloud compute ssh \
             --project {notebook_instance.project_id} \
             --zone {notebook_instance.zone} \
             --tunnel-through-iap \
              {notebook_instance.name} \
             -- -L 8080:localhost:{local_port}"
        if run_in_background:
            bash_command += " -N -f"

        process = subprocess.Popen(bash_command.split())
        process.communicate()

    def ssh(self, instance_name: str, run_in_background: bool, local_port: int) -> None:
        """
        A wrapper around _ssh method, but this function also verifies if the notebook is
        defined in YAML config and already started.

        Args:
            instance_name: name of the notebook to connect to
            run_in_background: whether to run in the background or in interactive mode
            local_port: jupyter lab will be exposed at this port at localhost

        Returns:

        """
        if instance_name == "all":
            if len(self.config.notebooks) == 1:
                self._ssh(self.config.notebooks[0], run_in_background, local_port)
            elif len(self.config.notebooks) > 1:
                raise ValueError("You can connect to only one notebook at a time.")
            else:
                logger.user_error("No notebook definition found in your YAML config.")
        else:
            if instance_name in [notebook.name for notebook in self.config.notebooks]:
                self._ssh(
                    [notebook for notebook in self.config.notebooks if notebook.name == instance_name][0],
                    run_in_background,
                    local_port,
                )
            else:
                logger.user_error(f"No notebook {instance_name} found")
                raise ValueError(f"Notebook {instance_name} does not exists in configuration")

    def build(self) -> int:
        for instance in self.instances:
            self._create_instance_request(notebook_instance=instance, deploy=False)
        logger.user_success("Notebooks validation OK!")
        return 0



import logging
from enum import Enum
from pathlib import Path

import typer
from cookiecutter.main import cookiecutter

from wanna.core.loggers.wanna_logger import get_logger

from .plugins.runner import PluginRunner
from .version import perform_check

logger = get_logger(__name__)

logging.getLogger("smart_open").setLevel(logging.ERROR)
logging.getLogger("google").setLevel(logging.ERROR)
logging.getLogger("kfp").setLevel(logging.ERROR)

runner = PluginRunner()
app = runner.app


class WannaRepositoryTemplate(str, Enum):
    sklearn = "sklearn"
    blank = "blank"


@app.command(name="version", help="Print your current and latest available version")
def version(ctx: typer.Context):
    perform_check()


@app.command(name="init", help="Initiate a new wanna-ml project from template.")
def init(
    ctx: typer.Context,
    output_dir: Path = typer.Option(
        ...,
        "--output-dir",
        prompt="Where do you want to initiate your wanna-ml repository? (input '.' to use the current directory)",
        help="The output directory where wanna-ml repository will be created",
    ),
    template: WannaRepositoryTemplate = typer.Option(
        WannaRepositoryTemplate.sklearn.value,
        "--template",
        "-t",
        help="Choose from available repository templates",
        show_choices=True,
    ),
):
    repository_templates = {
        WannaRepositoryTemplate.sklearn.value: "https://github.com/avast/wanna-ml-cookiecutter",
        WannaRepositoryTemplate.blank.value: "https://github.com/avast/wanna-blank-cookiecutter",
    }
    repository_template_url = repository_templates.get(template)
    result_dir = cookiecutter(repository_template_url, output_dir=output_dir)
    logger.user_success(f"Repo initiated at {result_dir}")


def wanna():
    """
    Main entrypoint for wanna cli
    """
    app()


if __name__ == "__main__":
    wanna()



from typing import List, Optional

from pydantic import BaseModel, EmailStr, Extra, Field, root_validator, validator

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.gcp_components import GPU, Disk, VMImage
from wanna.core.utils import validators


class BucketMount(BaseModel, extra=Extra.forbid):
    bucket_name: str
    mount_path: str = "/gcs"

    _bucket_name = validator("bucket_name")(validators.validate_bucket_name)


class NotebookEnvironment(BaseModel, extra=Extra.forbid):
    vm_image: Optional[VMImage]
    docker_image_ref: Optional[str]

    _ = root_validator()(validators.validate_only_one_must_be_set)


class NotebookModel(BaseInstanceModel):
    name: str = Field(min_length=3, max_length=63, to_lower=True, regex="^[a-z][a-z0-9-]*[a-z0-9]$")
    zone: str
    machine_type: str = "n1-standard-4"
    environment: NotebookEnvironment = NotebookEnvironment(vm_image=VMImage(framework="common", version="cpu"))
    instance_owner: Optional[EmailStr]
    gpu: Optional[GPU]
    boot_disk: Optional[Disk]
    data_disk: Optional[Disk]
    bucket_mounts: Optional[List[BucketMount]]
    network: Optional[str]
    subnet: Optional[str]
    tensorboard_ref: Optional[str]
    no_public_ip: bool = True
    no_proxy_access: bool = False

    _machine_type = validator("machine_type")(validators.validate_machine_type)


class ManagedNotebookModel(BaseInstanceModel):
    name: str = Field(min_length=3, max_length=63, to_lower=True, regex="^[a-z][a-z0-9-]*[a-z0-9]$")
    owner: str
    machine_type: Optional[str] = "n1-standard-4"
    gpu: Optional[GPU]
    data_disk: Optional[Disk]
    kernels: Optional[List[str]]
    tensorboard_ref: Optional[str]
    network: Optional[str]
    subnet: Optional[str]
    internal_ip_only: Optional[bool] = True
    idle_shutdown: Optional[bool]
    idle_shutdown_timeout: Optional[int] = Field(ge=10, le=1440)



import typer

from wanna.core.deployment.models import PushMode

profile_name_option = typer.Option(
    "default",
    "--profile",
    "-p",
    envvar="WANNA_GCP_PROFILE_NAME",
    help="Name of the GCP profile you want to use. "
    "Profiles are loaded from wanna-ml yaml config and "
    "(optionally) from this file too.",
)

wanna_file_option = typer.Option(
    "wanna.yaml", "--file", "-f", envvar="WANNA_FILE", help="Path to the wanna-ml yaml configuration"
)

push_mode_option: PushMode = typer.Option(
    PushMode.all,
    "--mode",
    "-m",
    help="Pipeline push mode, due to CI/CD not "
    "allowing to push to docker registry from "
    "GCP Agent, we need to split it. "
    "Use all for dev",
)


def instance_name_option(instance_type: str, operation: str, help: str = None):
    return typer.Option(
        "all",
        "--name",
        "-n",
        help=help
        or f"Specify only one {instance_type} from your wanna-ml yaml configuration to {operation}. "
        f"Choose 'all' to {operation} all {instance_type}s.",
    )



import os
import sys
from pathlib import Path

from jinja2 import Template


def render_template(source_path: Path, **kwargs) -> str:
    templates_dir = Path(os.path.dirname(sys.modules["wanna.core"].__file__)) / "templates"  # type: ignore
    source_path = templates_dir / source_path
    with open(source_path) as f:
        template = Template(f.read())
    return template.render(**kwargs)



import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from typing import List

from pydantic import BaseModel, EmailStr, Extra


class NotificationChannelModel(BaseModel, extra=Extra.forbid):
    name: str
    type: Literal["email"]
    emails: List[EmailStr]



from pydantic import Field

from wanna.core.models.base_instance import BaseInstanceModel


class TensorboardModel(BaseInstanceModel):
    name: str = Field(min_length=3, max_length=128)
    region: str



from typing import Optional

from pydantic import BaseModel, validator

from wanna.core.utils import validators


class CloudSchedulerModel(BaseModel):
    cron: str
    timezone: str = "Etc/UTC"
    service_account: Optional[str]

    # Validators
    _schedule = validator("cron")(validators.validate_cron_schedule)



import contextlib
import json
from pathlib import Path
from typing import Any, Dict, Union

from google.cloud.storage import Client
from smart_open import open as gcs_open

from wanna.core.deployment.credentials import GCPCredentialsMixIn


class IOMixin(GCPCredentialsMixIn):
    @contextlib.contextmanager
    def _open(self, uri, mode="r", **kwargs):
        transport_params = {"client": Client(credentials=self.credentials)} if str(uri).startswith("gs") else {}
        with gcs_open(uri, mode, transport_params=transport_params, **kwargs) as c:
            yield c

    def upload_file(self, source: str, destination: str):
        with self._open(source, "rb") as f:
            with self._open(destination, "wb") as fout:
                fout.write(f.read())

    def write(self, destination: Union[Path, str], body: str) -> None:
        with self._open(destination, "w") as fout:
            fout.write(body)

    def read(self, source: Union[Path, str]) -> Dict[Any, Any]:
        with self._open(source, "r") as fin:
            return json.loads(fin.read())



import abc
from typing import Callable, List

import typer


class BasePlugin(abc.ABC):
    """Some doc"""

    def __init__(self) -> None:
        self.app = typer.Typer()

    def register_command(self, func: Callable[..., None]) -> None:
        self.app.command()(func)

    def register_many(self, funcs: List[Callable[..., None]]) -> None:
        for func in funcs:
            self.register_command(func)



#!/usr/bin/env python

from distutils.core import setup

from setuptools import find_packages

setup(
    name="{{ cookiecutter.component_slug }}",
    version="{{ cookiecutter.component_version }}",
    description="{{ cookiecutter.component_description }}",
    author="{{ cookiecutter.component_author }}",
    author_email="{{ cookiecutter.component_author_email }}",
    url="{{ cookiecutter.component_url }}",
    python_requires=">=3.8",
    install_requires=[
        "click==8.1.3",
        "kfp==1.8.12",
        "google-cloud-aiplatform==1.13.0",
        "smart-open==6.0.0",
    ],
    include_package_data=True,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
)



from pathlib import Path
from typing import List, Tuple

import typer
from google.api_core import exceptions
from google.cloud.notebooks_v1.services.managed_notebook_service import ManagedNotebookServiceClient
from google.cloud.notebooks_v1.types import (
    CreateRuntimeRequest,
    EncryptionConfig,
    LocalDisk,
    LocalDiskInitializeParams,
    Runtime,
    RuntimeAcceleratorConfig,
    RuntimeAccessConfig,
    RuntimeSoftwareConfig,
    VirtualMachine,
    VirtualMachineConfig,
)
from waiting import wait

from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.notebook import ManagedNotebookModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils import templates
from wanna.core.utils.gcp import upload_string_to_gcs

logger = get_logger(__name__)


class ManagedNotebookService(BaseService[ManagedNotebookModel]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        version: str = "dev",
    ):
        super().__init__(
            instance_type="managed-notebook",
        )
        self.version = version
        self.instances = config.managed_notebooks
        self.notebook_client = ManagedNotebookServiceClient()
        self.config = config
        self.bucket_name = config.gcp_profile.bucket
        self.tensorboard_service = TensorboardService(config=config)

    def _create_runtime_request(self, instance: ManagedNotebookModel, deploy: bool = True) -> CreateRuntimeRequest:
        """
        Transform the information about desired managed-notebook from the model based on yaml config
        to the form suitable for GCP API.

        Args:
            instance

        Returns:
            CreateRuntimeRequest
        """
        # Configuration of the managed notebook
        labels = {
            "wanna_name": instance.name,
            "wanna_resource": self.instance_type,
        }
        if instance.labels:
            labels = {**instance.labels, **labels}
        # Disks
        disk_type = instance.data_disk.disk_type if instance.data_disk else None
        disk_size_gb = instance.data_disk.size_gb if instance.data_disk else None
        localDiskParams = LocalDiskInitializeParams(disk_size_gb=disk_size_gb, disk_type=disk_type)
        localDisk = LocalDisk(initialize_params=localDiskParams)
        encryption_config = (
            EncryptionConfig(kms_key=self.config.gcp_profile.kms_key) if self.config.gcp_profile.kms_key else None
        )

        # Accelerator
        if instance.gpu:
            runtimeAcceleratorConfig = RuntimeAcceleratorConfig(
                type=instance.gpu.accelerator_type, core_count=instance.gpu.count
            )
        else:
            runtimeAcceleratorConfig = None
        # Network and subnetwork
        network = instance.network if instance.network else self.config.gcp_profile.network
        full_network = f"projects/{instance.project_id}/global/networks/{network}"
        subnet = instance.subnet if instance.subnet else self.config.gcp_profile.subnet
        full_subnet = f"projects/{instance.project_id}/regions/{instance.region}/subnetworks/{subnet}"

        # Post startup script
        if deploy and instance.tensorboard_ref:
            script = self._prepare_startup_script(self.instances[0])
            blob = upload_string_to_gcs(
                script,
                instance.bucket or self.bucket_name,
                f"notebooks/{instance.name}/startup_script.sh",
            )
            post_startup_script = f"gs://{blob.bucket.name}/{blob.name}"
        else:
            post_startup_script = None

        # VM
        virtualMachineConfig = VirtualMachineConfig(
            machine_type=instance.machine_type,
            data_disk=localDisk,
            encryption_config=encryption_config,
            labels=labels,
            accelerator_config=runtimeAcceleratorConfig,
            network=full_network,
            subnet=full_subnet,
            internal_ip_only=instance.internal_ip_only,
            tags=instance.tags,
            metadata=instance.metadata,
        )
        virtualMachine = VirtualMachine(virtual_machine_config=virtualMachineConfig)
        # Runtime
        runtimeSoftwareConfig = RuntimeSoftwareConfig(
            {
                "kernels": instance.kernels,
                "post_startup_script": post_startup_script,
                "idle_shutdown": instance.idle_shutdown,
                "idle_shutdown_timeout": instance.idle_shutdown_timeout,
            }
        )
        runtimeAccessConfig = RuntimeAccessConfig(
            access_type=RuntimeAccessConfig.RuntimeAccessType.SINGLE_USER, runtime_owner=instance.owner
        )
        runtime = Runtime(
            access_config=runtimeAccessConfig, software_config=runtimeSoftwareConfig, virtual_machine=virtualMachine
        )
        # Create runtime request
        return CreateRuntimeRequest(
            parent=f"projects/{instance.project_id}/locations/{instance.region}",
            runtime_id=instance.name,
            runtime=runtime,
        )

    def _delete_one_instance(self, notebook_instance: ManagedNotebookModel) -> None:
        """
        Delete one notebook instance. This assumes that it has been already verified that notebook exists.

        Args:
            notebook_instance: notebook to delete
        """

        exists = self._instance_exists(notebook_instance)
        if exists:
            with logger.user_spinner(f"Deleting {self.instance_type} {notebook_instance.name}"):
                deleted = self.notebook_client.delete_runtime(
                    name=f"projects/{notebook_instance.project_id}/locations/"
                    f"{notebook_instance.region}/runtimes/{notebook_instance.name}"
                )
                deleted.result()
        else:
            logger.user_error(
                f"Notebook with name {notebook_instance.name} was not found in region {notebook_instance.region}",
            )

    def _create_one_instance(self, instance: ManagedNotebookModel, **kwargs) -> None:
        """
        Create a notebook instance based on information in NotebookModel class.
        1. Check if the notebook already exists
        2. Parse the information from NotebookModel to GCP API friendly format = runtime_request
        3. Wait for the compute instance behind the notebook to start
        4. Wait for JupyterLab to start
        5. Get and print the link to JupyterLab

        Args:
            instance: notebook to be created

        """
        exists = self._instance_exists(instance)
        if exists:
            logger.user_info(f"Managed notebook {instance.name} already exists in location {instance.region}")
            should_recreate = typer.confirm("Are you sure you want to delete it and start a new?")
            if should_recreate:
                self._delete_one_instance(instance)
            else:
                return

        request = self._create_runtime_request(instance=instance, deploy=True)
        with logger.user_spinner(f"Creating instance for {instance.name}"):
            nb_instance = self.notebook_client.create_runtime(request=request)
            instance_full_name = (
                nb_instance.result().name
            )  # .result() waits for compute engine behind the notebook to start
        with logger.user_spinner("Starting JupyterLab"):
            wait(
                lambda: self._validate_jupyterlab_state(instance_full_name, Runtime.State.ACTIVE),
                timeout_seconds=450,
                sleep_seconds=20,
                waiting_for="Starting JupyterLab in your instance",
            )
            jupyterlab_link = self._get_jupyterlab_link(instance_full_name)
        logger.user_success(f"JupyterLab started at {jupyterlab_link}")

    def _list_running_instances(self, project_id: str, location: str) -> List[str]:
        """
        List all notebooks with given project_id and location.

        Args:
            project_id: GCP project ID
            location: GCP location (zone)

        Returns:
            instance_names: List of the full names on notebook instances (this includes project_id, and zone)

        """
        instances = self.notebook_client.list_runtimes(parent=f"projects/{project_id}/locations/{location}")
        instance_names = [i.name for i in instances.runtimes]
        return instance_names

    def _instance_exists(self, instance: ManagedNotebookModel) -> bool:
        """
        Check if the instance with given instance_name exists in given GCP project project_id and location.
        Args:
            instance: notebook to verify if exists on GCP

        Returns:
            True if exists, False if not
        """
        full_instance_name = f"projects/{instance.project_id}/locations/{instance.region}/runtimes/{instance.name}"
        return full_instance_name in self._list_running_instances(
            instance.project_id, instance.region or self.config.gcp_profile.region
        )

    def _prepare_startup_script(self, nb_instance: ManagedNotebookModel) -> str:
        """
        Prepare the notebook startup script based on the information from notebook.
        This script run at the Compute Engine Instance creation time with a root user.

        Args:
            nb_instance

        Returns:
            startup_script
        """
        if nb_instance.tensorboard_ref:
            tensorboard_resource_name = self.tensorboard_service.get_or_create_tensorboard_instance_by_name(
                nb_instance.tensorboard_ref
            )
        else:
            tensorboard_resource_name = None
        startup_script = templates.render_template(
            Path("notebook_startup_script.sh.j2"),
            tensorboard_resource_name=tensorboard_resource_name,
        )
        return startup_script

    def _validate_jupyterlab_state(self, instance_id: str, state: int) -> bool:
        """
        Validate if the given notebook instance is in given state.

        Args:
            instance_id: Full notebook instance id
            state: Notebook state (ACTIVE, PENDING,...)

        Returns:
            True if desired state, False otherwise
        """
        try:
            instance_info = self.notebook_client.get_runtime(name=instance_id)
        except exceptions.NotFound:
            raise exceptions.NotFound(f"Managed Notebook {instance_id} was not found.") from None
        return instance_info.state == state

    def _get_jupyterlab_link(self, instance_id: str) -> str:
        """
        Get a link to jupyterlab proxy based on given notebook instance id.
        Args:
            instance_id: full notebook instance id

        Returns:
            proxy_uri: link to jupyterlab
        """
        instance_info = self.notebook_client.get_runtime({"name": instance_id})
        return f"https://{instance_info.access_config.proxy_uri}"

    def _return_diff(self) -> Tuple[List[str], List[ManagedNotebookModel]]:
        """
        Figuring out the diff between GCP and wanna.yaml. Lists managed notebooks to be deleted and created.
        """
        parent = f"projects/{self.config.gcp_profile.project_id}/locations/{self.config.gcp_profile.region}"
        active_runtimes = self.notebook_client.list_runtimes(parent=parent).runtimes
        wanna_names = [managednotebook.name for managednotebook in self.instances]
        existing_names = [str(runtime.name) for runtime in active_runtimes]
        to_be_deleted = []
        to_be_created = []
        """
        Notebooks to be deleted
        """
        for runtime in active_runtimes:
            if runtime.virtual_machine.virtual_machine_config.labels["wanna_project"] == self.config.wanna_project.name:
                if runtime.name.split("/")[-1] not in wanna_names:
                    to_be_deleted.append(str(runtime.name))
        """
        Notebooks to be created
        """
        for notebook in self.instances:
            if (
                f"projects/{notebook.project_id}/locations/{notebook.region}/runtimes/{notebook.name}"
                not in existing_names
            ):
                to_be_created.append(notebook)

        return to_be_deleted, to_be_created

    def sync(self, force) -> None:
        """
        1. Reads current notebooks where label is defined per field wanna_project.name in wanna.yaml
        2. Does a diff between what is on GCP and what is on yaml
        3. Delete the ones in GCP that are not in wanna.yaml
        4. Create the ones defined in yaml and missing in GCP
        """
        to_be_deleted, to_be_created = self._return_diff()

        if to_be_deleted:
            to_be_deleted_str = "\n".join(["- " + item for item in to_be_deleted])
            logger.user_info(f"Managed notebooks to be deleted:\n{to_be_deleted_str}")
            should_delete = True if force else typer.confirm("Are you sure you want to delete them?")
            if should_delete:
                for item in to_be_deleted:
                    with logger.user_spinner(f"Deleting {item}"):
                        self.notebook_client.delete_runtime(name=item).result()

        if to_be_created:
            to_be_created_str = "\n".join(["- " + item.name for item in to_be_created])
            logger.user_info(f"Managed notebooks to be created:\n{to_be_created_str}")
            should_create = True if force else typer.confirm("Are you sure you want to create them?")
            if should_create:
                # mypy inspects item is as str which obviously isn't hence the type: ignore
                for item in to_be_created:  # type: ignore
                    self._create_one_instance(item)  # type: ignore
                    logger.user_info(f"Created {item.name}")  # type: ignore

        logger.user_info("Managed notebooks on GCP are in sync with wanna.yaml")

    def build(self) -> int:
        for instance in self.instances:
            self._create_runtime_request(instance=instance, deploy=False)
        logger.user_success("Managed notebooks validation OK!")
        return 0



from caseconverter import snakecase


def remove_nones(d):
    """
    Delete keys with the value ``None`` or `null` in a dictionary, recursively.

    """

    for key, value in list(d.items()):
        if value is None:
            del d[key]
        elif isinstance(value, dict):
            remove_nones(value)
        elif isinstance(value, list):
            for v in value:
                if isinstance(v, dict):
                    remove_nones(v)
        else:
            del d[key]
            d[snakecase(key)] = value
    return d



from typing import List, Optional

from pydantic import BaseModel, EmailStr, Extra


class WannaProjectModel(BaseModel, extra=Extra.forbid):
    name: str
    version: str
    authors: List[EmailStr]
    billing_id: Optional[str]
    organization_id: Optional[str]



from abc import ABC
from typing import Generic, List, Optional, TypeVar

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.utils.gcp import convert_project_id_to_project_number, get_network_info

logger = get_logger(__name__)

T = TypeVar("T", bound=BaseInstanceModel)


class BaseService(ABC, Generic[T]):
    """
    This is a base service and every other service (notebooks, jobs, pipelines,...)
    should build on top of this. It implements some basic function (create, delete)
    and utils for loading the wanna config.

    Args:
        instance_type: what are you working with (notebook, job, tensorboard) - used mainly in logging
        instance_model: pydantic instance model
    """

    def __init__(
        self,
        instance_type: str,
    ):
        self.instances: List[T] = []
        self.instance_type = instance_type

    def create(self, instance_name: str, **kwargs) -> None:
        """
        Create an instance with name "name" based on wanna-ml config.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be created.
                  Set to "all" to create everything from wanna-ml yaml configuration.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            self._create_one_instance(instance, **kwargs)

    def delete(self, instance_name: str) -> None:
        """
        Delete an instance with name "name" based on wanna-ml config if exists on GCP.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be deleted.
                  Set to "all" to create all from configuration.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            self._delete_one_instance(instance)

    def stop(self, instance_name: str) -> None:
        """
        Stop an instance with name "name" based on wanna-ml config.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be stopped.
                  Set to "all" to delete everything from wanna-ml yaml configuration for this resource.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            self._stop_one_instance(instance)

    def _delete_one_instance(self, instance: T) -> None:
        """
        Abstract class. Should delete one instance based on one model (eg. delete one notebook).
        """
        ...

    def _create_one_instance(self, instance: T, **kwargs) -> None:
        """
        Abstract class. Should create one instance based on one model (eg. create one notebook).
        """
        ...

    def _stop_one_instance(self, instance: T) -> None:
        """
        Abstract class. Should stop one instance based on one model (eg. stop one job).
        """
        ...

    def _filter_instances_by_name(self, instance_name: str) -> List[T]:
        """
        From self.instances filter only the instances with name instance_name.

        Args:
            instance_name: Name of the instance to return. Set to "all" to return all instances.

        Returns:
            instances

        """
        if instance_name == "all":
            instances = self.instances
            if not instances:
                logger.user_error(f"No {self.instance_type} can be parsed from your wanna-ml yaml config.")
        else:
            instances = [nb for nb in self.instances if nb.name == instance_name]
        if not instances:
            logger.user_error(
                f"{self.instance_type} with name {instance_name} not found in your wanna-ml yaml config.",
            )

        return instances

    def report(
        self,
        instance_name: str,
        wanna_project: str,
        wanna_resource: str,
        gcp_project: str,
        billing_id: Optional[str],
        organization_id: Optional[str],
    ) -> None:
        """
        Sends a link to cost report
        Billing and Organization IDs are hard coded
        """
        base_url = f"https://console.cloud.google.com/billing/{billing_id}/reports;projects={gcp_project}"
        organization = f"?organizationId={organization_id}"

        if instance_name == "all":
            labels = f";labels=wanna_project:{wanna_project},wanna_resource:{wanna_resource}"
        elif instance_name not in [nb.name for nb in self.instances]:
            logger.user_error(
                f"{self.instance_type} with name {instance_name} not found in your wanna-ml yaml config.",
            )
            return
        else:
            labels = f";labels=wanna_project:{wanna_project},wanna_resource:{wanna_resource},wanna_name:{instance_name}"

        link = base_url + labels + organization
        logger.user_info(f"Here is a link to your {wanna_resource} cost report:")
        logger.user_success(f"{link}")

    def _get_resource_network(
        self,
        project_id: str,
        push_mode: PushMode,
        resource_network: Optional[str],
        fallback_project_network: Optional[str],
        use_project_number: bool = True,
    ):
        resource_network = resource_network if resource_network else fallback_project_network
        if resource_network:
            result = get_network_info(resource_network)
            if result:
                # long format network found
                project_id, resource_network = result

            # In certain scenarios we can't build on GCP and have no access to GCP from within Avast build infra
            if push_mode.can_push_gcp_resources():
                if not project_id.isdigit() and use_project_number:
                    project_id = convert_project_id_to_project_number(project_id)

            return f"projects/{project_id}/global/networks/{resource_network}"
        else:
            return None

    def _get_resource_subnet(self, network: Optional[str], subnet: Optional[str], region: Optional[str]):
        if subnet:
            # Assumes the full qualified path was provided in config
            if "/" in subnet:
                return subnet
            else:
                # Get the project id from the provided network and pass it to subnet
                if network:
                    network_project_parts = network.split("/")
                    if len(network_project_parts) >= 2:
                        network_project_id = network_project_parts[1]
                    else:
                        network_project_id = network
                return f"projects/{network_project_id}/regions/{region}/subnetworks/{subnet}"
        else:
            return None



import os
from pathlib import Path
from typing import Union

import kfp.components as comp


def load_wanna_component(path: Union[Path, str]):
    with open(str(path), "r") as f:
        component = os.path.expandvars(f.read())
        return comp.load_component_from_text(component)



import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from enum import Enum
from pathlib import Path
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

from pydantic import BaseModel, EmailStr, Extra
from pydantic.generics import GenericModel

from wanna.core.models.cloud_scheduler import CloudSchedulerModel
from wanna.core.models.docker import DockerBuildResult
from wanna.core.models.notification_channel import NotificationChannelModel
from wanna.core.models.training_custom_job import BaseCustomJobModel


class GCPResource(GenericModel, extra=Extra.forbid, validate_assignment=True, arbitrary_types_allowed=True):
    name: str
    project: str
    location: str
    service_account: Optional[EmailStr]

    def get_base_resource(self):
        return self.dict()


class NotificationChannelResource(GCPResource):
    type_: Literal["email"]
    config: Dict[str, str]
    labels: Dict[str, str]


class CloudSchedulerResource(GCPResource):
    cloud_scheduler: CloudSchedulerModel
    body: Dict[str, Any]
    labels: Dict[str, str]
    notification_channels: List[str]


class CloudFunctionResource(GCPResource):
    build_dir: Path
    resource_root: str
    resource_function_template: str
    resource_requirements_template: str
    template_vars: Dict[str, Any]
    env_params: Dict[str, str]
    labels: Dict[str, str]
    network: str
    notification_channels: List[str]


class LogMetricResource(GCPResource):
    filter_: str
    description: str


class AlertPolicyResource(GCPResource):
    logging_metric_type: str
    resource_type: str
    display_name: str
    labels: Dict[str, str]
    notification_channels: List[str]


class PipelineResource(GCPResource):
    pipeline_name: str
    pipeline_bucket: str
    pipeline_root: str
    pipeline_version: str
    json_spec_path: str
    parameter_values: Dict[str, Any] = {}
    labels: Dict[str, str] = {}
    enable_caching: bool = True
    schedule: Optional[CloudSchedulerModel]
    docker_refs: List[DockerBuildResult]
    compile_env_params: Dict[str, str]
    network: str
    notification_channels: List[NotificationChannelModel] = []
    encryption_spec_key_name: Optional[str]


JOB = TypeVar("JOB", bound=BaseCustomJobModel, covariant=True)  # dependency from wanna models


class JobResource(GCPResource, Generic[JOB]):
    job_payload: Dict[str, Any]
    image_refs: List[str] = []
    tensorboard: Optional[str]
    network: Optional[str]
    job_config: JOB
    encryption_spec_key_name: Optional[str]


class PushArtifact(BaseModel):
    name: str


class JsonArtifact(PushArtifact):
    name: str
    json_body: Dict[Any, Any]
    destination: str


class PathArtifact(PushArtifact):
    name: str
    source: str
    destination: str


class ContainerArtifact(PushArtifact):
    name: str
    tags: List[str]


class PushMode(str, Enum):
    all = "all"
    manifests = "manifests"
    containers = "containers"
    quick = "quick"

    def is_quick_mode(self) -> bool:
        return self == PushMode.manifests or self == PushMode.quick

    def can_push_containers(self) -> bool:
        return self == PushMode.all or self == PushMode.containers

    def can_push_gcp_resources(self) -> bool:
        return self == PushMode.all or self == PushMode.manifests


class PushTask(BaseModel):
    manifest_artifacts: List[PathArtifact]
    container_artifacts: List[ContainerArtifact]
    json_artifacts: List[JsonArtifact]


PushResult = List[Tuple[List[ContainerArtifact], List[PathArtifact], List[JsonArtifact]]]



import json
import os
from pathlib import Path
from typing import List, Optional, Tuple

from caseconverter import snakecase
from google.cloud import aiplatform
from kfp.v2.compiler.main import compile_pyfile
from python_on_whales import Image

from wanna.core.deployment.artifacts_push import PushResult
from wanna.core.deployment.models import (
    ContainerArtifact,
    JsonArtifact,
    PathArtifact,
    PipelineResource,
    PushMode,
    PushTask,
)
from wanna.core.deployment.vertex_connector import VertexConnector
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.docker import DockerBuildResult, DockerImageModel, ImageBuildType
from wanna.core.models.pipeline import PipelineModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.services.base import BaseService
from wanna.core.services.docker import DockerService
from wanna.core.services.path_utils import PipelinePaths
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils.loaders import load_yaml_path

logger = get_logger(__name__)


class PipelineService(BaseService[PipelineModel]):
    def __init__(
        self,
        config: WannaConfigModel,
        workdir: Path,
        version: str = "dev",
        push_mode: PushMode = PushMode.all,
        connector: VertexConnector[PipelineResource] = VertexConnector[PipelineResource](),
    ):
        super().__init__(
            instance_type="pipeline",
        )
        self.connector = connector
        self.instances = config.pipelines
        self.config = config
        self.workdir = workdir
        self.tensorboard_service = TensorboardService(config=config)
        self.version = version
        self.push_mode = push_mode
        self.docker_service = DockerService(
            docker_model=config.docker,  # type: ignore
            gcp_profile=config.gcp_profile,
            version=version,
            work_dir=workdir,
            wanna_project_name=self.config.wanna_project.name,
            quick_mode=push_mode.is_quick_mode(),
        )
        self.notification_channels = {channel.name: channel for channel in self.config.notification_channels}

    def build(self, instance_name: str) -> List[Path]:
        """
        Create an instance with name "name" based on wanna-ml config.
        Args:
            instance_name: The name of the only instance from wanna-ml config that should be created.
                  Set to "all" to create everything from wanna-ml yaml configuration.
        """
        instances = self._filter_instances_by_name(instance_name)
        return [self._compile_one_instance(instance) for instance in instances]

    def push(self, manifests: List[Path], local: bool = False) -> PushResult:
        return self.connector.push_artifacts(
            self.docker_service.push_image,
            self._prepare_push(manifests, self.version, local),
        )

    def _prepare_push(self, pipelines: List[Path], version: str, local: bool = False) -> List[PushTask]:
        push_tasks = []
        for local_manifest_path in pipelines:
            manifest = PipelineService.read_manifest(self.connector, str(local_manifest_path))
            pipeline_paths = PipelinePaths(self.workdir, manifest.pipeline_bucket, manifest.pipeline_name)
            json_artifacts, manifest_artifacts, container_artifacts = [], [], []

            logger.user_info(text=f"Packaging {manifest.pipeline_name} pipeline resources")

            # Push containers if we are running on Internal Teamcity build agent or on all push-mode
            if self.push_mode.can_push_containers():
                for ref in manifest.docker_refs:
                    if ref.build_type != ImageBuildType.provided_image:
                        container_artifacts.append(ContainerArtifact(name=ref.name, tags=ref.tags))

            # Push gcp resources if we are running on GCP build agent
            if self.push_mode.can_push_gcp_resources():
                # Prepare manifest paths
                local_kubeflow_json_spec_path = pipeline_paths.get_local_pipeline_json_spec_path(version)
                wanna_manifest_publish_path = pipeline_paths.get_gcs_wanna_manifest_path(version)
                kubeflow_json_spec_publish_path = pipeline_paths.get_gcs_pipeline_json_spec_path(version)

                if local:
                    # Override paths to local dir when in "local" mode, IE tests or local run
                    wanna_manifest_publish_path = pipeline_paths.get_local_wanna_manifest_path(version)
                    kubeflow_json_spec_publish_path = pipeline_paths.get_local_pipeline_json_spec_path(version)

                # Ensure to update manifest json_spec_path to have the actual gcs location
                manifest.json_spec_path = kubeflow_json_spec_publish_path

                json_artifacts.append(
                    JsonArtifact(
                        name="WANNA pipeline manifest",
                        json_body=manifest.dict(),
                        destination=wanna_manifest_publish_path,
                    )
                )

                manifest_artifacts.append(
                    PathArtifact(
                        name="Kubeflow V2 pipeline spec",
                        source=local_kubeflow_json_spec_path,
                        destination=manifest.json_spec_path,
                    )
                )

            push_tasks.append(
                PushTask(
                    manifest_artifacts=manifest_artifacts,
                    container_artifacts=container_artifacts,
                    json_artifacts=json_artifacts,
                )
            )

        return push_tasks

    @staticmethod
    def get_pipeline_bucket(bucket: Optional[str], fallback_bucket: str):
        if bucket:
            return bucket if bucket.startswith("gs://") else f"gs://{bucket}"
        else:
            return f"gs://{fallback_bucket}"

    def deploy(self, instance_name: str, env: str):
        instances = self._filter_instances_by_name(instance_name)
        for pipeline in instances:
            logger.user_info(f"Deploying {pipeline.name} version {self.version} to env {env}")
            pipeline_bucket = PipelineService.get_pipeline_bucket(pipeline.bucket, self.config.gcp_profile.bucket)
            pipeline_paths = PipelinePaths(self.workdir, pipeline_bucket, pipeline_name=pipeline.name)
            manifest = PipelineService.read_manifest(
                self.connector, pipeline_paths.get_gcs_wanna_manifest_path(self.version)
            )
            self.connector.deploy_pipeline(manifest, pipeline_paths, self.version, env)

    @staticmethod
    def run(
        pipelines: List[str],
        extra_params: Optional[Path] = None,
        sync: bool = True,
    ) -> None:
        connector = VertexConnector[PipelineResource]()
        for manifest_path in pipelines:
            manifest = PipelineService.read_manifest(connector, str(manifest_path))
            aiplatform.init(location=manifest.location, project=manifest.project)
            connector.run_pipeline(manifest, extra_params, sync)

    def _export_pipeline_params(
        self,
        pipeline_paths: PipelinePaths,
        pipeline_instance: PipelineModel,
        version: str,
        images: List[Tuple[DockerImageModel, Optional[Image], str]],
        tensorboard: Optional[str],
        network: Optional[str],
    ):
        # Prepare env params to be exported
        pipeline_env_params = {
            "project_id": pipeline_instance.project_id,
            "pipeline_name": pipeline_instance.name,
            "version": version,
            "bucket": pipeline_instance.bucket,
            "region": pipeline_instance.region,
            "pipeline_root": pipeline_paths.get_gcs_pipeline_root(),
            "pipeline_labels": json.dumps(pipeline_instance.labels),
            "pipeline_service_account": (
                pipeline_instance.service_account
                if pipeline_instance.service_account
                else self.config.gcp_profile.service_account
            ),
        }

        if self.config.gcp_profile.kms_key:
            pipeline_env_params["encryption_spec_key_name"] = self.config.gcp_profile.kms_key

        if tensorboard:
            pipeline_env_params["tensorboard"] = tensorboard

        if network:
            pipeline_env_params["pipeline_network"] = network

        # Export Pipeline wanna ENV params to be available during compilation
        pipeline_name_prefix = snakecase(f"{pipeline_instance.name}").upper()
        for key, value in pipeline_env_params.items():
            env_name = snakecase(f"{pipeline_name_prefix}_{key.upper()}").upper()
            os.environ[env_name] = value

        for (docker_image_model, _, tag) in images:
            env_name = snakecase(f"{docker_image_model.name}_DOCKER_URI").upper()
            os.environ[env_name] = tag

        # Collect pipeline compile params from wanna config
        if pipeline_instance.pipeline_params and isinstance(pipeline_instance.pipeline_params, Path):
            pipeline_params_path = (self.workdir / pipeline_instance.pipeline_params).resolve()
            pipeline_compile_params = load_yaml_path(pipeline_params_path, self.workdir)
        elif pipeline_instance.pipeline_params and isinstance(pipeline_instance.pipeline_params, dict):
            pipeline_compile_params = pipeline_instance.pipeline_params
        else:
            pipeline_compile_params = {}

        return pipeline_env_params, pipeline_compile_params

    def _compile_one_instance(self, pipeline: PipelineModel) -> Path:

        image_tags = [
            self.docker_service.get_image(docker_image_ref=docker_image_ref)
            for docker_image_ref in pipeline.docker_image_ref
        ]

        logger.user_info(text=f"Compiling pipeline {pipeline.name}")

        # Prep build dir
        pipeline_bucket = PipelineService.get_pipeline_bucket(pipeline.bucket, self.config.gcp_profile.bucket)
        pipeline_paths = PipelinePaths(self.workdir, pipeline_bucket, pipeline.name)

        tensorboard = (
            self.tensorboard_service.get_or_create_tensorboard_instance_by_name(pipeline.tensorboard_ref)
            if pipeline.tensorboard_ref and self.push_mode.can_push_gcp_resources()
            else None
        )

        network = self._get_resource_network(
            project_id=pipeline.project_id,
            push_mode=self.push_mode,
            resource_network=pipeline.network,
            fallback_project_network=self.config.gcp_profile.network,
        )

        labels = {
            "wanna_resource_name": pipeline.name,
            "wanna_resource": self.instance_type,
        }
        if pipeline.sla_hours:
            labels["wanna_sla_hours"] = str(pipeline.sla_hours).replace(".", "_")
        if pipeline.labels:
            labels = {**pipeline.labels, **labels}
        pipeline.labels = labels
        # Collect kubeflow pipeline params for compilation
        pipeline_env_params, pipeline_params = self._export_pipeline_params(
            pipeline_paths, pipeline, self.version, image_tags, tensorboard, network
        )

        # Compile kubeflow V2 Pipeline
        compile_pyfile(
            pyfile=str(self.workdir / pipeline.pipeline_file),
            function_name=pipeline.pipeline_function,
            pipeline_parameters=pipeline_params,
            package_path=pipeline_paths.get_local_pipeline_json_spec_path(self.version),
            type_check=True,
            use_experimental=False,
        )

        docker_refs = [
            DockerBuildResult(
                name=model.name,
                tags=image.repo_tags if image and image.repo_tags else [tag],
                build_type=model.build_type,
            )
            for model, image, tag in image_tags
        ]

        channels = []
        for ref in pipeline.notification_channels_ref:
            channel = self.notification_channels.get(ref)
            if channel:
                channels.append(channel)
            else:
                raise ValueError(f"{ref} notification channel not specified")

        deployment_manifest = PipelineResource(
            name=f"pipeline {pipeline.name}",
            project=pipeline.project_id,
            location=pipeline.region,
            service_account=pipeline.service_account,
            network=network,
            pipeline_name=pipeline.name,
            pipeline_bucket=pipeline_bucket,
            pipeline_version=self.version,
            json_spec_path=pipeline_paths.get_local_pipeline_json_spec_path(self.version),
            parameter_values=pipeline_params,
            enable_caching=True,
            labels=pipeline.labels,
            pipeline_root=pipeline_env_params.get("pipeline_root"),
            schedule=pipeline.schedule,
            docker_refs=docker_refs,
            compile_env_params=pipeline_env_params,
            notification_channels=channels,
            encryption_spec_key_name=self.config.gcp_profile.kms_key,
        )

        manifest_path = pipeline_paths.get_local_wanna_manifest_path(self.version)
        self.connector.write(manifest_path, deployment_manifest.json())
        return Path(manifest_path).resolve()

    @staticmethod
    def read_manifest(connector: VertexConnector[PipelineResource], path: str) -> PipelineResource:
        return PipelineResource.parse_obj(connector.read(path))

    def _delete_one_instance(self, instance: PipelineModel) -> None:
        pass

    def _create_one_instance(self, instance: PipelineModel, **kwargs) -> None:
        pass

    def _instance_exists(self, instance: PipelineModel) -> bool:
        pass



import pathlib
from pathlib import Path
from typing import Optional

import typer

from wanna.cli.plugins.base_plugin import BasePlugin
from wanna.cli.plugins.common_options import instance_name_option, profile_name_option, wanna_file_option
from wanna.core.services.notebook import NotebookService
from wanna.core.utils.config_loader import load_config_from_yaml


class NotebookPlugin(BasePlugin):
    """
    Create, delete and more operations for user-managed Workbench (Jupyter notebook).
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_many(
            [
                self.delete,
                self.create,
                self.ssh,
                self.report,
                self.build,
            ]
        )

    @staticmethod
    def delete(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("notebook", "delete"),
    ) -> None:
        """
        Delete a User-Managed Workbench Notebook.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = NotebookService(config=config, workdir=workdir)
        nb_service.delete(instance_name)

    @staticmethod
    def create(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("notebook", "create"),
        owner: Optional[str] = typer.Option(None, "--owner", "-o", help=""),
    ) -> None:
        """
        Create a User-Managed Workbench Notebook.

        If there already is a notebook with the same name in the same location and project,
        you will be prompt if you want to delete the existing one and start a new one.

        When the notebook instance is created, you will be given a URL link to JupyterLab.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = NotebookService(config=config, workdir=workdir, owner=owner)
        nb_service.create(instance_name)

    @staticmethod
    def ssh(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option(
            "notebook",
            "ssh",
            help="Specify to which notebook you want to connect via ssh. "
            "Selecting 'all' will work only if there is just one notebook "
            "defined in your configuration, an error will be thrown otherwise.",
        ),
        run_in_background: bool = typer.Option(
            False,
            "--background/--interactive",
            "-b/-i",
            help="Interactive mode will start a bash directly in the Compute Engine instance "
            "backing the Jupyter notebook. "
            "Background mode serves more like a port-forwarding, "
            "you will be able to connect to the Jupyter Lab at localhost:{LOCAL_PORT}",
        ),
        local_port: int = typer.Option(
            8080, "--port", help="Jupyter Lab will be accessible at this port at localhost."
        ),
    ) -> None:
        """
        SSH connect to the Compute Engine instance that is behind the Jupyter Notebook.

        This will only work if the notebook is already running.

        Please note that you can connect to only one instance with one command call.
        If you have more notebooks defined in your YAML config, you have to select to which
        you want to connect to, instance_name "all" will be refused.

        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = NotebookService(config=config, workdir=workdir)
        nb_service.ssh(instance_name, run_in_background, local_port)

    @staticmethod
    def report(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("notebook", "report"),
    ) -> None:
        """
        Displays a link to the cost report per wanna_project and optionally per instance name.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = NotebookService(config=config, workdir=workdir)
        nb_service.report(
            instance_name=instance_name,
            wanna_project=config.wanna_project.name,
            wanna_resource="notebook",
            gcp_project=config.gcp_profile.project_id,
            billing_id=config.wanna_project.billing_id,
            organization_id=config.wanna_project.organization_id,
        )

    @staticmethod
    def build(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
    ) -> None:
        """
        Validates build of notebooks as they are defined in wanna.yaml
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = NotebookService(config=config, workdir=workdir)
        nb_service.build()



import logging
import re

from cron_validator import CronValidator
from google.api_core import exceptions
from google.cloud.notebooks_v1.types.instance import Instance
from google.cloud.storage import Client as StorageClient

from wanna.core.utils.credentials import get_credentials
from wanna.core.utils.gcp import (
    DEFAULT_VALIDATION_MODE,
    VALIDATION_MODE,
    get_available_compute_image_families,
    get_available_compute_machine_types,
    get_available_regions,
    get_available_zones,
    get_network_info,
)


def validate_zone(zone, values):
    available_zones = get_available_zones(project_id=values.get("project_id"))
    if zone not in available_zones:
        raise ValueError(f"Zone invalid ({zone}). must be on of: {available_zones}")
    return zone


def validate_region(region, values):
    available_regions = get_available_regions(project_id=values.get("project_id"))
    if region not in available_regions:
        raise ValueError(f"Region invalid ({region}). must be on of: {available_regions}")
    return region


def validate_machine_type(machine_type, values):
    available_machine_types = get_available_compute_machine_types(
        project_id=values.get("project_id"), zone=values.get("zone")
    )
    if machine_type not in available_machine_types:
        raise ValueError(f"Machine type invalid ({machine_type}). must be on of: {available_machine_types}")
    return machine_type


def validate_requirements(cls, v):
    if not any(v.values()):
        raise ValueError(
            "One of requirements.file (path to your requirements.txt) or "
            "requirements.package_list (list of pip packages to install) "
            "must be set if you want to install python packages with requirements block."
        )
    return v


def validate_network_name(network_name):
    if not get_network_info(network_name):
        if not re.match("^[a-z][a-z0-9-]+$", network_name):
            raise ValueError(
                "Invalid format of network name. Either use the full name of VPC network"
                "'projects/{project_id}/global/networks/{network_id}'"
                "or just '{network_id}'. In the second case, the project_id will be parsed from notebook settings."
            )
    return network_name


def validate_bucket_name(bucket_name):
    if VALIDATION_MODE == DEFAULT_VALIDATION_MODE:
        try:
            _ = StorageClient(credentials=get_credentials()).get_bucket(bucket_name)
        except exceptions.NotFound:
            raise ValueError(f"Bucket with name {bucket_name} does not exist")
        except exceptions.Forbidden:
            logging.warning(f"Your user does not have permission to access bucket {bucket_name}")

    return bucket_name


def validate_vm_image(cls, v):
    framework = v.get("framework")
    version = v.get("version")
    os = v.get("os")
    available_image_families = get_available_compute_image_families(
        project="deeplearning-platform-release",
        image_filter="(-deprecated:*)",
        family_must_contain="notebook",
    )
    available_frameworks = set(i.get("framework") for i in available_image_families)
    if framework not in available_frameworks:
        raise ValueError(f"VM Image framework {framework} not available. Choose one of: {available_frameworks}")

    available_versions = set(i.get("version") for i in available_image_families if i.get("framework") == framework)
    if version not in available_versions:
        raise ValueError(
            f"VM Image version {version} not available for {framework}. Choose one of: {available_versions}"
        )

    if os:
        available_os = set(
            i.get("os")
            for i in available_image_families
            if i.get("framework") == framework and i.get("version") == version
        )
        if os not in available_os:
            raise ValueError(
                f"VM Image OS {os} not available for {framework} and version {version}. Choose one of: {available_os}"
            )

    return v


def validate_only_one_must_be_set(cls, v):
    items_set = {key for key, value in v.items() if value is not None}
    if len(items_set) == 0:
        raise ValueError(f"One of {list(v.keys())} must be set.")
    elif len(items_set) > 1:
        raise ValueError(f"Specify only one of {items_set}")
    return v


def validate_cron_schedule(schedule: str):
    if schedule is not None and CronValidator.parse(schedule) is None:
        raise ValueError(f"Cron expression is invalid ({schedule}).")
    else:
        return schedule


def validate_disk_type(disk_type):
    disk_type = disk_type.upper()
    if disk_type not in Instance.DiskType.__members__:
        raise ValueError(f"Disk type invalid ({type}). must be on of: {Instance.DiskType._member_names_}")
    return disk_type


def validate_accelerator_type(
    accelerator_type: Instance.AcceleratorType,
) -> Instance.AcceleratorType:
    if accelerator_type not in Instance.AcceleratorType.__members__:
        raise ValueError(
            f"GPU accelerator type invalid ({accelerator_type})."
            f"must be on of: {Instance.AcceleratorType._member_names_}"
        )
    return accelerator_type


def validate_project_id(project_id: str) -> str:
    if not re.match("^[a-z][a-z0-9-]{4,28}[a-z0-9]$", project_id):
        raise ValueError(
            "Invalid GCP project id. project_id: "
            "Must be 6 to 30 characters in length. "
            "Contains lowercase letters, numbers, and hyphens. "
            "Must start with a letter. "
            "Cannot end with a hyphen."
        )
    return project_id



import os
from pathlib import Path

from caseconverter import kebabcase

from wanna.core.utils.gcp import is_gcs_path


class JobPaths:

    job_manifest_filename = "job-manifest.json"

    def __init__(self, workdir: Path, bucket: str, job_name: str):
        self.pipeline_name = job_name
        self.workdir = workdir
        self.bucket = bucket
        self.pipelines_dir = (workdir / "build").resolve()
        self.local_job_path = self._get_job_path(str(self.pipelines_dir), job_name)
        self.gcs_job_path = self._get_job_path(self.bucket, job_name)

    def _get_job_path(self, base_path: str, pipeline_name: str):
        if not is_gcs_path(base_path):
            os.makedirs(base_path, exist_ok=True)
        return f"{base_path}/wanna-jobs/{kebabcase(pipeline_name).lower()}"

    def _get_job_manifests_path(self, base_path: str, version: str):
        path = f"{base_path}/deployment/{version}/manifests"
        if not is_gcs_path(path):
            os.makedirs(path, exist_ok=True)
        return path

    def get_local_job_manifest_path(self, version: str):
        return self._get_job_manifests_path(self.local_job_path, version)

    def get_gcs_job_manifest_path(self, version: str):
        return self._get_job_manifests_path(self.gcs_job_path, version)

    def get_local_job_wanna_manifest_path(self, version: str):
        return f"{self.get_local_job_manifest_path(version)}/{self.job_manifest_filename}"

    def get_gcs_job_wanna_manifest_path(self, version: str):
        return f"{self.get_gcs_job_manifest_path(version)}/{self.job_manifest_filename}"


class PipelinePaths:

    json_spec_filename = "pipeline-spec.json"
    wanna_manifest_filename = "wanna-manifest.json"
    job_manifest_filename = "job-manifest.json"

    def __init__(self, workdir: Path, bucket: str, pipeline_name: str):
        self.pipeline_name = pipeline_name
        self.workdir = workdir
        self.bucket = bucket
        self.pipelines_dir = (workdir / "build").resolve()
        self.local_pipeline_path = self._get_pipeline_path(str(self.pipelines_dir), pipeline_name)
        self.gcs_pipeline_path = self._get_pipeline_path(self.bucket, pipeline_name)
        self.local_job_path = self._get_pipeline_path(str(self.pipelines_dir), pipeline_name)
        self.gcs_job_path = self._get_pipeline_path(self.bucket, pipeline_name)

    def _get_pipeline_path(self, base_path: str, pipeline_name: str):
        if not is_gcs_path(base_path):
            os.makedirs(base_path, exist_ok=True)
        return f"{base_path}/wanna-pipelines/{kebabcase(pipeline_name).lower()}"

    def _get_pipeline_deployment_path(self, base_path: str, version: str):
        path = f"{base_path}/deployment/{version}"
        if not is_gcs_path(path):
            os.makedirs(path, exist_ok=True)
        return path

    def _get_pipeline_manifests_path(self, base_path: str, version: str):
        path = f"{base_path}/deployment/{version}/manifests"
        if not is_gcs_path(path):
            os.makedirs(path, exist_ok=True)
        return path

    def get_local_pipeline_manifest_path(self, version: str):
        return self._get_pipeline_manifests_path(self.local_pipeline_path, version)

    def get_gcs_pipeline_manifest_path(self, version: str):
        return self._get_pipeline_manifests_path(self.gcs_pipeline_path, version)

    def get_local_pipeline_deployment_path(self, version: str):
        return self._get_pipeline_deployment_path(self.local_pipeline_path, version)

    def get_gcs_pipeline_deployment_path(self, version: str):
        return self._get_pipeline_deployment_path(self.gcs_pipeline_path, version)

    def get_local_pipeline_json_spec_path(self, version: str):
        return f"{self.get_local_pipeline_manifest_path(version)}/{self.json_spec_filename}"

    def get_gcs_pipeline_json_spec_path(self, version: str):
        return f"{self.get_gcs_pipeline_manifest_path(version)}/{self.json_spec_filename}"

    def get_local_wanna_manifest_path(self, version: str):
        return f"{self.get_local_pipeline_manifest_path(version)}/{self.wanna_manifest_filename}"

    def get_gcs_wanna_manifest_path(self, version: str):
        return f"{self.get_gcs_pipeline_manifest_path(version)}/{self.wanna_manifest_filename}"

    def get_gcs_pipeline_root(self):
        return f"{self.gcs_pipeline_path}/executions/"

    def get_local_pipeline_root(self):
        return f"{self.local_pipeline_path}/executions/"



import os
import pathlib
from pathlib import Path
from typing import Any, Dict

from google.cloud import aiplatform

from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.models.wanna_config import WannaConfigModel
from wanna.core.utils import loaders
from wanna.core.utils.credentials import get_credentials

logger = get_logger(__name__)


def load_gcp_profile(profile_name: str, wanna_dict: Dict[str, Any]) -> GCPProfileModel:
    """
    This functions goes through wanna-ml config and optionally through a file
    $WANNA_GCP_PROFILE_PATH, reads all gcp profiles and returns the one
    with "profile_name" name
    Args:
        profile_name: name of the GCP profile
        wanna_dict: wanna-ml configuration as a dictionary
    Returns:
        GCPProfileModel

    """
    profiles = wanna_dict.get("gcp_profiles", {})
    profiles = {p.get("profile_name"): p for p in profiles}

    extra_profiles_path = os.getenv("WANNA_GCP_PROFILE_PATH")
    if extra_profiles_path and os.path.isfile(extra_profiles_path):
        extra_profiles = loaders.load_yaml_path(
            Path(extra_profiles_path), Path(extra_profiles_path).parent.absolute()
        ).get("gcp_profiles")
        if extra_profiles:
            extra_profiles = {p.get("profile_name"): p for p in extra_profiles}
            profiles.update(extra_profiles)

    if profile_name not in profiles:
        raise ValueError(f"Profile {profile_name} not found")

    selected_profile = profiles.get(profile_name)
    profile_model = GCPProfileModel.parse_obj(selected_profile)
    return profile_model


def load_config_from_yaml(wanna_config_path: Path, gcp_profile_name: str) -> WannaConfigModel:
    """
    Load the yaml file from wanna_config_path and parses the information to the models.
    This also includes the data validation.

    Args:
        wanna_config_path: path to the wanna-ml yaml file
        gcp_profile_name: name of the GCP profile
        profiles_file_path: optionally passed through cli

    Returns:
        WannaConfigModel

    """

    with logger.user_spinner("Reading and validating wanna yaml config"):
        with open(wanna_config_path) as file:
            # Load workflow file
            wanna_dict = loaders.load_yaml(file, pathlib.Path(wanna_config_path).parent.resolve())
        profile_model = load_gcp_profile(profile_name=gcp_profile_name, wanna_dict=wanna_dict)
        os.environ["GOOGLE_CLOUD_PROJECT"] = profile_model.project_id
        wanna_dict.update({"gcp_profile": profile_model})
        del wanna_dict["gcp_profiles"]
        wanna_config = WannaConfigModel.parse_obj(wanna_dict)
    logger.user_info(f"GCP profile '{profile_model.profile_name}' will be used.")
    logger.user_info(f"Profile details: {profile_model}")

    aiplatform.init(
        project=wanna_config.gcp_profile.project_id,
        location=wanna_config.gcp_profile.region,
        credentials=get_credentials(),
    )

    return wanna_config



import sys
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from pydantic import BaseModel, Extra, Field


class DockerBuildConfigModel(BaseModel, extra=Extra.forbid):
    # Docu for more info: https://gabrieldemarmiesse.github.io/python-on-whales/sub-commands/buildx/#build
    build_args: Dict[str, str] = {}
    add_hosts: Dict[str, str] = {}
    labels: Dict[str, str] = {}
    network: Optional[str]
    platforms: Optional[List[str]]
    secrets: Union[str, List[str]] = []
    ssh: Optional[str]
    target: Optional[str]


class ImageBuildType(str, Enum):
    local_build_image = "local_build_image"
    provided_image = "provided_image"
    notebook_ready_image = "notebook_ready_image"


class BaseDockerImageModel(BaseModel, extra=Extra.forbid, validate_assignment=True):
    name: str = Field(min_length=3, max_length=128)


class LocalBuildImageModel(BaseDockerImageModel):
    build_type: Literal[ImageBuildType.local_build_image]
    build_args: Optional[Dict[str, str]]
    context_dir: Path
    dockerfile: Path


class ProvidedImageModel(BaseDockerImageModel):
    build_type: Literal[ImageBuildType.provided_image]
    image_url: str


class NotebookReadyImageModel(BaseDockerImageModel):
    build_type: Literal[ImageBuildType.notebook_ready_image]
    build_args: Optional[Dict[str, str]]
    base_image: str = "gcr.io/deeplearning-platform-release/base-cpu"
    requirements_txt: Path


DockerImageModel = Union[LocalBuildImageModel, ProvidedImageModel, NotebookReadyImageModel]


class DockerModel(BaseModel, extra=Extra.forbid, validate_assignment=True):
    images: List[DockerImageModel] = []
    repository: Optional[str]
    registry: Optional[str]
    cloud_build: bool = False


class DockerBuildResult(BaseModel, extra=Extra.forbid):
    name: str
    tags: List[str]
    build_type: ImageBuildType



import functools
import os
from typing import Optional

from google import auth
from google.auth import impersonated_credentials
from google.auth.credentials import Credentials
from google.auth.exceptions import DefaultCredentialsError

from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)


@functools.lru_cache(maxsize=1)
def get_credentials() -> Optional[Credentials]:
    impersonate_account = os.getenv("WANNA_IMPERSONATE_ACCOUNT")
    target_scopes = ["https://www.googleapis.com/auth/cloud-platform"]

    try:
        if impersonate_account:
            _credentials, _ = auth.default()
            credentials = impersonated_credentials.Credentials(
                source_credentials=_credentials,
                target_principal=impersonate_account,
                target_scopes=_credentials.scopes if _credentials and _credentials.scopes else target_scopes,
                lifetime=500,
            )
            return credentials
        else:
            return None

    except DefaultCredentialsError as e:
        logger.user_info(
            f"{e}\ndefault credentials were not found you likely need to execute"
            "`gcloud auth application-default login`"
        )
        exit(1)



import os
import re
import tarfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import google.auth
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import storage
from google.cloud.compute import MachineTypesClient, ZonesClient
from google.cloud.compute_v1 import RegionsClient
from google.cloud.compute_v1.services.images import ImagesClient
from google.cloud.compute_v1.types import ListImagesRequest
from google.cloud.resourcemanager_v3.services.projects import ProjectsClient

from wanna.core.utils.credentials import get_credentials

DEFAULT_VALIDATION_MODE = "remote"
VALIDATION_MODE = os.environ.get("WANNA_VALIDATION_MODE", DEFAULT_VALIDATION_MODE)
NETWORK_REGEX = (
    "projects/((?:(?:[-a-z0-9]{1,63}\\.)*(?:[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?):)"
    "?(?:[0-9]{1,19}|(?:[a-z0-9](?:[-a-z0-9]{0,61}[a-z0-9])?)))/global/networks/"
    "((?:[a-z](?:[-a-z0-9]*[a-z0-9])?))$"
)


def are_gcp_credentials_set() -> bool:
    """
    Function to verify if the default GCP credentials can be abstracted
    from environment.

    Returns:
        True if GCP credentials can be found, False otherwise
    """
    try:
        _credentials, _project_id = google.auth.default()
        return True
    except DefaultCredentialsError:
        return False


def get_current_local_gcp_project_id() -> str:
    """
    Get current local GCP default project.

    Returns:
        project_id: your current local GCP default project
    """
    _, project_id = google.auth.default()
    return project_id


def get_available_compute_machine_types(project_id: str, zone: str) -> List[str]:
    """
    Get available GCP Compute Engine Machine Types based on project and zone.
    Args:
        project_id: GCP project id
        zone: GCP project location (zone)

    Returns:
        list of available machine types
    """
    # TODO: remove once we can push to teamcity via GCP
    if VALIDATION_MODE == DEFAULT_VALIDATION_MODE:
        response = MachineTypesClient(credentials=get_credentials()).list(project=project_id, zone=zone)
        machine_types = [mtype.name for mtype in response.items]
    else:
        machine_types = [
            "c2-standard-30",
            "c2-standard-4",
            "c2-standard-60",
            "c2-standard-8",
            "e2-highcpu-16",
            "e2-highcpu-2",
            "e2-highcpu-32",
            "e2-highcpu-4",
            "e2-highcpu-8",
            "e2-highmem-16",
            "e2-highmem-2",
            "e2-highmem-4",
            "e2-highmem-8",
            "e2-medium",
            "e2-micro",
            "e2-small",
            "e2-standard-16",
            "e2-standard-2",
            "e2-standard-32",
            "e2-standard-4",
            "e2-standard-8",
            "f1-micro",
            "g1-small",
            "m1-megamem-96",
            "m1-ultramem-160",
            "m1-ultramem-40",
            "m1-ultramem-80",
            "m2-megamem-416",
            "m2-ultramem-208",
            "m2-ultramem-416",
            "n1-highcpu-16",
            "n1-highcpu-2",
            "n1-highcpu-32",
            "n1-highcpu-4",
            "n1-highcpu-64",
            "n1-highcpu-8",
            "n1-highcpu-96",
            "n1-highmem-16",
            "n1-highmem-2",
            "n1-highmem-32",
            "n1-highmem-4",
            "n1-highmem-64",
            "n1-highmem-8",
            "n1-highmem-96",
            "n1-megamem-96",
            "n1-standard-1",
            "n1-standard-16",
            "n1-standard-2",
            "n1-standard-32",
            "n1-standard-4",
            "n1-standard-64",
            "n1-standard-8",
            "n1-standard-96",
            "n1-ultramem-160",
            "n1-ultramem-40",
            "n1-ultramem-80",
            "n2-highcpu-16",
            "n2-highcpu-2",
            "n2-highcpu-32",
            "n2-highcpu-4",
            "n2-highcpu-48",
            "n2-highcpu-64",
            "n2-highcpu-8",
            "n2-highcpu-80",
            "n2-highcpu-96",
            "n2-highmem-128",
            "n2-highmem-16",
            "n2-highmem-2",
            "n2-highmem-32",
            "n2-highmem-4",
            "n2-highmem-48",
            "n2-highmem-64",
            "n2-highmem-8",
            "n2-highmem-80",
            "n2-highmem-96",
            "n2-standard-128",
            "n2-standard-16",
            "n2-standard-2",
            "n2-standard-32",
            "n2-standard-4",
            "n2-standard-48",
            "n2-standard-64",
            "n2-standard-8",
            "n2-standard-80",
            "n2-standard-96",
            "n2d-highcpu-128",
            "n2d-highcpu-16",
            "n2d-highcpu-2",
            "n2d-highcpu-224",
            "n2d-highcpu-32",
            "n2d-highcpu-4",
            "n2d-highcpu-48",
            "n2d-highcpu-64",
            "n2d-highcpu-8",
            "n2d-highcpu-80",
            "n2d-highcpu-96",
            "n2d-highmem-16",
            "n2d-highmem-2",
            "n2d-highmem-32",
            "n2d-highmem-4",
            "n2d-highmem-48",
            "n2d-highmem-64",
            "n2d-highmem-8",
            "n2d-highmem-80",
            "n2d-highmem-96",
            "n2d-standard-128",
            "n2d-standard-16",
            "n2d-standard-2",
            "n2d-standard-224",
            "n2d-standard-32",
            "n2d-standard-4",
            "n2d-standard-48",
            "n2d-standard-64",
            "n2d-standard-8",
            "n2d-standard-80",
            "n2d-standard-96",
            "t2d-standard-1",
            "t2d-standard-16",
            "t2d-standard-2",
            "t2d-standard-32",
            "t2d-standard-4",
            "t2d-standard-48",
            "t2d-standard-60",
            "t2d-standard-8",
        ]

    return machine_types


def get_available_zones(project_id: str) -> List[str]:
    """
    Get available GCP zones based on project.
    Args:
        project_id: GCP project id

    Returns:
        list of available zones
    """

    # TODO: remove once we can push to teamcity via GCP
    if VALIDATION_MODE == DEFAULT_VALIDATION_MODE:
        response = ZonesClient(credentials=get_credentials()).list(project=project_id)
        return [zone.name for zone in response.items]
    else:
        return [
            "us-east1-c",
            "us-east1-d",
            "us-west1-b",
            "us-west1-c",
            "us-west1-a",
            "europe-west1-b",
            "europe-west1-d",
            "europe-west1-c",
            "europe-west3-c",
            "europe-west3-a",
            "europe-west3-b",
        ]


def get_available_regions(project_id: str) -> List[str]:
    """
    Get available GCP regions based on project.
    Args:
        project_id: GCP project id

    Returns:
        list of available regions
    """
    # TODO: enable once we can push to teamcity via GCP
    if VALIDATION_MODE == DEFAULT_VALIDATION_MODE:
        response = RegionsClient(credentials=get_credentials()).list(project=project_id)
        return [region.name for region in response.items]
    else:
        return ["europe-west1", "europe-west3", "us-east1", "us-west1"]


def get_region_from_zone(zone: str) -> str:
    """
    Get available GCP region from zone.
    Args:
        zone: GCP zone

    Returns:
        region: GCP region
    """
    return zone.rpartition("-")[0]


def convert_project_id_to_project_number(project_id: str) -> str:
    """
    Convert GCP project_id (eg. 'my-gcp-project') to project_number (eg. '966193337054')

    Args:
        project_id: GCP project id

    Returns:
        project_number: GCP project number
    """
    project_name = ProjectsClient(credentials=get_credentials()).get_project(name=f"projects/{project_id}").name
    project_number = re.sub("projects/", "", project_name)
    return project_number


def parse_image_name_family(name: str) -> Dict[str, Any]:
    """
    Based on GCP Compute Engine VM Image name family (eg. tf2-2-7-cu113-notebooks-debian-10)
    return framework (eg. tf2), version (eg. 2-7-cu113), os (debian-10) information.

    Args:
        name: VM Image family name

    Returns:
        Dictionary with framework, version and os

    """
    framework = name.partition("-")[0]
    match_version = re.search("(?<=-)(.*?)(?=-notebooks)", name)
    assert match_version is not None  # TODO: improve result assert with message
    version = match_version.group()
    os = None if name.endswith("notebooks") else name.split("-notebooks-")[-1]
    return {"framework": framework, "version": version, "os": os}


def get_available_compute_image_families(
    project: str, image_filter: str = None, family_must_contain: str = None
) -> List[Dict[str, str]]:
    """
    List available Compute Engine VM image families.

    Args:
        project: VM Image project ID
        image_filter: filter for the images https://googleapis.dev/python/compute/latest/compute_v1/types.html#google.cloud.compute_v1.types.ListImagesRequest.filter
        family_must_contain: additional string that must be a part of the image family name for easier filtering
                                (eg. notebook to filter only the Vertex AI Workbench notebook-ready images)

    Returns:
        List of dicts from parse_image_name_family

    """  # noqa: E501
    list_images_request = ListImagesRequest(project=project, filter=image_filter)
    all_images = ImagesClient(credentials=get_credentials()).list(list_images_request)
    if family_must_contain:
        return [parse_image_name_family(image.family) for image in all_images if family_must_contain in image.family]
    return [parse_image_name_family(image.family) for image in all_images]


def construct_vm_image_family_from_vm_image(framework: str, version: str, os: Optional[str]) -> str:
    """
    Construct name of the Compute Engine VM family with given framework(eg. pytorch),
    version(eg. 1-9-xla) and optional OS (eg. debian-10).

    Args:
        framework: VM image framework (pytorch, r, tf2, ...)
        version: Version of the framework
        os: operation system

    Returns:
        object: Compute Engine VM Family name
    """
    if os:
        return f"{framework}-{version}-notebooks-{os}"
    return f"{framework}-{version}-notebooks"


def make_tarfile(source_dir: Path, output_filename: Path):
    """
    TAR a given folder and save the result to output_filename.

    Args:
        source_dir:
        output_filename:
    """
    os.makedirs(output_filename.parent.absolute(), exist_ok=True)
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname=".")


def upload_file_to_gcs(filename: Path, bucket_name: str, blob_name: str) -> storage.blob.Blob:
    """
    Upload file to GCS bucket

    Args:
        filename: local file
        bucket_name:
        blob_name:

    Returns:
        storage.blob.Blob
    """
    storage_client = storage.Client(credentials=get_credentials())
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)
    blob.upload_from_filename(filename)
    return blob


def upload_string_to_gcs(data: str, bucket_name: str, blob_name: str) -> storage.blob.Blob:
    """
    Upload a string to GCS bucket without saving it locally as a file.
    Args:
        data: string that will form a file on GCS
        bucket_name:
        blob_name:

    Returns:
        storage.blob.Blob
    """
    storage_client = storage.Client(credentials=get_credentials())
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)
    blob.upload_from_string(data)
    return blob


def is_gcs_path(path: str):
    """
    Simply checks if path str represents a GCS path
    Args:
        path: path string that will be checked

    Returns:
        bool
    """
    return path.startswith("gs://")


def get_network_info(network: Optional[str]) -> Optional[Tuple[str, str]]:
    """
    gets information about a network if set in long format
    Args:
        network: string to extract network info from
        bucket_name:
        blob_name:

    Returns:
        Tuple[str, str]
    """
    if network:
        result = re.search(NETWORK_REGEX, network)
        if result:
            return result.group(1), result.group(2)

    return None



from pathlib import Path
from typing import Any, Dict, TextIO

import yaml
from yamlinclude import YamlIncludeConstructor


def load_yaml(stream: TextIO, context_dir: Path, **extras: Any) -> Dict[Any, Any]:
    """
    Convert a YAML stream into a class via the OrderedLoader class.
    """
    YamlIncludeConstructor.add_to_loader_class(loader_class=yaml.FullLoader, base_dir=context_dir)
    yaml_dict = yaml.load(stream, Loader=yaml.FullLoader) or {}
    yaml_dict.update(extras)
    return yaml_dict


def load_yaml_path(path: Path, context_dir: Path, **extras: Any) -> Dict[Any, Any]:
    """
    Convert a Path into a yaml dict
    """
    with open(path, "r") as f:
        return load_yaml(f, context_dir, **extras)



import sys
from typing import Any, Dict, List, Optional, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from pydantic import BaseModel, Extra, Field, root_validator, validator
from typing_extensions import Annotated

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.gcp_components import GPU, Disk


class PythonPackageModel(BaseModel, extra=Extra.forbid):
    docker_image_ref: str
    package_gcs_uri: str
    module_name: str


class ContainerModel(BaseModel, extra=Extra.forbid):
    docker_image_ref: str
    command: Optional[List[str]]


class WorkerPoolModel(BaseModel, extra=Extra.forbid):
    python_package: Optional[PythonPackageModel]
    container: Optional[ContainerModel]
    args: Optional[List[Union[str, float, int]]]
    env: Optional[Dict[str, str]]
    machine_type: str = "n1-standard-4"
    gpu: Optional[GPU]
    boot_disk: Optional[Disk]
    replica_count: int = 1

    # _machine_type = validator("machine_type", allow_reuse=True)(validators.validate_machine_type)

    @root_validator
    def one_from_python_or_container_spec_must_be_set(cls, values):  # pylint: disable=no-self-argument,no-self-use
        if values.get("python_package") and values.get("container"):
            raise ValueError("Only one of python_package or container can be set")
        if not values.get("python_package") and not values.get("container"):
            raise ValueError("At least one of python_package or container must be set")
        return values


class ReductionServerModel(BaseModel, extra=Extra.forbid):
    replica_count: int
    machine_type: str
    container_uri: str


class IntegerParameter(BaseModel, extra=Extra.forbid):
    type: Literal["integer"]
    var_name: str
    min: int
    max: int
    scale: Literal["log", "linear"] = "linear"


class DoubleParameter(BaseModel, extra=Extra.forbid):
    type: Literal["double"]
    var_name: str
    min: float
    max: float
    scale: Literal["log", "linear"] = "linear"


class CategoricalParameter(BaseModel, extra=Extra.forbid):
    type: Literal["categorical"]
    var_name: str
    values: List[str]


class DiscreteParameter(BaseModel, extra=Extra.forbid):
    type: Literal["discrete"]
    var_name: str
    scale: Literal["log", "linear"] = "linear"
    values: List[int]


HyperParamater = Annotated[
    Union[IntegerParameter, DoubleParameter, CategoricalParameter, DiscreteParameter], Field(discriminator="type")
]


class HyperparameterTuning(BaseModel):
    metrics: Dict[str, Literal["minimize", "maximize"]]
    parameters: List[HyperParamater]
    max_trial_count: int = 15
    parallel_trial_count: int = 3
    search_algorithm: Optional[Literal["grid", "random"]]
    encryption_spec_key_name: Optional[str]


class BaseCustomJobModel(BaseInstanceModel):
    region: str
    enable_web_access: bool = False
    bucket: str
    base_output_directory: Optional[str]
    tensorboard_ref: Optional[str]
    timeout_seconds: int = 60 * 60 * 24  # 24 hours
    encryption_spec_key_name: Optional[str]

    @root_validator(pre=False)
    def _set_base_output_directory_if_not_provided(  # pylint: disable=no-self-argument,no-self-use
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        if not values.get("base_output_directory"):
            values["base_output_directory"] = f"gs://{values.get('bucket')}/wanna-jobs/{values.get('name')}/outputs"
        return values

    @root_validator(pre=False)
    def _service_account_must_be_set_when_using_tensorboard(  # pylint: disable=no-self-argument,no-self-use
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        if values.get("tensorboard_ref") and not values.get("service_account"):
            raise ValueError("service_account must be set when using tensorboard in jobs")
        return values


# https://cloud.google.com/vertex-ai/docs/training/create-custom-job
class CustomJobModel(BaseCustomJobModel):
    workers: List[WorkerPoolModel]
    hp_tuning: Optional[HyperparameterTuning]

    @validator("workers", pre=False)
    def _worker_pool_must_have_same_spec(  # pylint: disable=no-self-argument,no-self-use
        cls, workers: List[WorkerPoolModel]
    ) -> List[WorkerPoolModel]:
        if workers:
            python_packages = list(filter(lambda w: w.python_package is not None, workers))
            containers = list(filter(lambda w: w.container is not None, workers))
            if len(python_packages) > 0 and len(containers) > 0:
                raise ValueError(
                    "CustomJobs must be of the same spec. " "Either just based on python_package or container"
                )

        return workers


# https://cloud.google.com/vertex-ai/docs/training/create-training-pipeline
class TrainingCustomJobModel(BaseCustomJobModel):
    worker: WorkerPoolModel
    reduction_server: Optional[ReductionServerModel]



from pathlib import Path
from typing import Optional

import typer

from wanna.cli.plugins.base_plugin import BasePlugin
from wanna.cli.plugins.common_options import instance_name_option, profile_name_option, wanna_file_option
from wanna.core.services.tensorboard import TensorboardService
from wanna.core.utils.config_loader import load_config_from_yaml


class TensorboardPlugin(BasePlugin):
    """
    Create, delete or list Tensorboard instances.
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_many(
            [
                self.delete,
                self.create,
                self.list,
            ]
        )

        # add some nesting with `sub-notebook-command` command.
        # self.app.add_typer(SubNotebookPlugin().app, name='sub-notebook-command')

    @staticmethod
    def delete(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("tensorboard", "delete"),
    ) -> None:
        """
        Delete Tensorboard Instance in GCP Vertex AI Experiments.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        tb_service = TensorboardService(config=config)
        tb_service.delete(instance_name)

    @staticmethod
    def create(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("tensorboard", "create"),
    ) -> None:
        """
        Create Tensorboard Instance in GCP Vertex AI Experiments.

        If there already is a tensorboard with the same name in the same location and project,
        you will be prompt if you want to delete the existing one and start a new one.

        When the tensorboard instance is created, you will be given a full resource name.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        tb_service = TensorboardService(config=config)
        tb_service.create(instance_name)

    @staticmethod
    def list(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        region: Optional[str] = typer.Option(
            None, "--region", help="Overwrites the region from wanna-ml yaml configuration"
        ),
        filter_expr: str = typer.Option(
            None,
            "--filter",
            help="GCP filter expression for tensorboard instances. "
            "Read more on GCP filters on "
            "https://cloud.google.com/sdk/gcloud/reference/topic/filters \n"
            "Example: display_name=my-tensorboard. \n"
            "Example: labels.wanna_project:* - to show all tensorboard created by wanna-ml.\n"
            "Example: labels.wanna_project:sushi-ssl.",
        ),
        show_url: bool = typer.Option(True, "--url/--no-url", help="Weather to show URL link to experiments"),
    ) -> None:
        """
        List Tensorboard Instances in GCP Vertex AI Experiments.

        We also show Tensorboard Experiments and Tensorboard Runs for each Instance
        in the tree format.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        tb_service = TensorboardService(config=config)
        region = region or config.gcp_profile.region
        if not region:
            raise ValueError(
                "Please provide a region. Either via cli arg or via region or zone in selected gcp profile"
            )
        else:
            tb_service.list_tensorboards_in_tree(region=region, filter_expr=filter_expr, show_url=show_url)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wanna',
 'wanna.cli',
 'wanna.cli.plugins',
 'wanna.components',
 'wanna.components.kubeflow',
 'wanna.components.templates',
 'wanna.components.templates.base.{{ cookiecutter.component_slug }}',
 'wanna.components.templates.base.{{ cookiecutter.component_slug }}.src.{{ '
 'cookiecutter.component_slug }}',
 'wanna.components.templates.base.{{ cookiecutter.component_slug }}.tests',
 'wanna.core',
 'wanna.core.deployment',
 'wanna.core.loggers',
 'wanna.core.models',
 'wanna.core.services',
 'wanna.core.templates',
 'wanna.core.utils']

package_data = \
{'': ['*'], 'wanna.components.templates': ['base/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=5.4.0,<6.0.0',
 'case-converter>=1.0.2,<2.0.0',
 'checksumdir>=1.2.0,<2.0.0',
 'cookiecutter>=2.1.1,<3.0.0',
 'cron-validator>=1.0.6,<2.0.0',
 'email-validator>=1.2.1,<2.0.0',
 'emoji>=1.7.0,<2.0.0',
 'gitpython>=3.1.27,<4.0.0',
 'google-api-core>=2.7.3,<3.0.0',
 'google-cloud-aiplatform>=1.13.0,<2.0.0',
 'google-cloud-build>=3.8.2,<4.0.0',
 'google-cloud-compute>=1.3.0,<2.0.0',
 'google-cloud-functions>=1.6.0,<2.0.0',
 'google-cloud-logging>=3.1.2,<4.0.0',
 'google-cloud-monitoring>=2.9.2,<3.0.0',
 'google-cloud-notebooks>=1.3.0,<2.0.0',
 'google-cloud-pipeline-components>=1.0.6,<2.0.0',
 'google-cloud-resource-manager>=1.4.1,<2.0.0',
 'google-cloud-scheduler>=2.6.3,<3.0.0',
 'google-cloud-storage>=1.44,<2.0',
 'halo>=0.0.31,<0.0.32',
 'importlib-metadata>=4.0,<5.0',
 'kfp>=1.8.12,<2.0.0',
 'packaging>=21.3,<22.0',
 'pathvalidate>=2.5.0,<3.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-on-whales>=0.43.0,<0.44.0',
 'pyyaml-include>=1.3,<2.0',
 'rich>=12.5.1,<13.0.0',
 'smart-open[gcs]>=6.0,<7.0',
 'treelib>=1.6.1,<2.0.0',
 'typer==0.6.1',
 'types-setuptools>=63.2.3,<64.0.0',
 'waiting>=1.4.1,<2.0.0']

entry_points = \
{'console_scripts': ['wanna = wanna.cli.__main__:wanna']}

setup_kwargs = {
    'name': 'wanna-ml',
    'version': '0.2.7',
    'description': 'CLI tool for managing ML projects on Vertex AI',
    'long_description': '# WANNA-ML\n\n---\n\n<p align="center" font-style="italic"> \n<em> Complete MLOps framework for Vertex-AI  </em>\n</p>\n\n---\n\n<p align="center">\n<a href="https://github.com/avast/wanna-ml/actions/workflows/build.yml" target="_blank">\n    <img src="https://github.com/avast/wanna-ml/actions/workflows/build.yml/badge.svg" alt="Build">\n</a>\n<a href="https://github.com/avast/wanna-ml/actions/workflows/release.yml" target="_blank">\n    <img src="https://github.com/avast/wanna-ml/actions/workflows/release.yml/badge.svg" alt="Release">\n</a>\n<a href="https://codecov.io/gh/avast/wanna-ml" target="_blank">\n    <img src="https://codecov.io/gh/avast/wanna-ml/branch/master/graph/badge.svg?token=TAFWK4GJPR" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/wanna-ml/" target="_blank">\n    <img src="https://img.shields.io/pypi/v/wanna-ml?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n</p>\n\n# About WANNA-ML\n\nWANNA-ML is a CLI tool that helps researchers, data scientists, and ML Engineers quickly adapt to Google Cloud Platform (GCP) and get started on the cloud in almost no time.\n\nIt makes it easy to start a Jupyter notebook, run training jobs and pipelines, build a Docker container, export logs to Tensorboards, and much more.\n\nWe build on top of Vertex-AI managed services and integrate with other GCP services like Cloud Build and Artifact Registry to provide you with a standardized structure for managing ML assets on GCP.\n\n\n## Help\n\nSee the [documentation](https://avast.github.io/wanna-ml/) for more details.\n\n\n## Get started\n\n### Installation\nInstall using `pip install -U wanna-ml`.\n\nFor more information on the installation process and requirements, visit out [installation page in documentation](https://avast.github.io/wanna-ml/installation)\n\n### Use Docker Image\nFor your convenience, we have prepared a Docker image with everything you need to get started.\n```bash\ndocker pull michalmrazek9/wanna-ml\n\ndocker run -it michalmrazek9/wanna-ml\n\n$ wanna version\n```\n\n### Authentication\nWANNA-ML relies on `gcloud` for user authentication. \n\n1. Install the `gcloud` CLI - follow [official guide](https://cloud.google.com/sdk/docs/install)\n2. Authenticate with the `gcloud init`\n3. Set you Google Application Credentials `gcloud auth application-default login`\n\n### Docker Build\nYou can use a local Docker daemon to build Docker images, but it is not required. \nYou are free to choose between local building on GCP Cloud Build. \nIf you prefer local Docker image building, install  [Docker Desktop](https://www.docker.com/products/docker-desktop/).\n\n### GCP IAM Roles and Permissions\nDifferent WANNA-ML calls require different GCP permissions to create given resources on GCP. Our [documentation page](https://avast.github.io/wanna-ml/)\nlists recommended GCP IAM roles for each `wanna` command.\n\n## Examples\nJump to [the samples](https://github.com/avast/wanna-ml/tree/master/samples) to see a complete solution \nfor various use cases.\n\n## Issues\nPlease report issues to [GitHub](https://github.com/avast/wanna-ml/issues).\n\n## Contributing\nYour contributions are always welcome, see [CONTRIBUTING.md](https://github.com/avast/wanna-ml/blob/master/CONTRIBUTING.md) for more information.\nIf you like WANNA-ML, don\'t forget to give our project a star! \n\n## Licence\nDistributed under the MIT License - see [LICENSE](https://github.com/avast/wanna-ml/blob/master/LICENCE).\n',
    'author': 'Joao Da Silva',
    'author_email': 'joao.silva1@avast.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://avast.github.io/wanna-ml',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import Field

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.cloud_scheduler import CloudSchedulerModel


class PipelineModel(BaseInstanceModel):
    name: str = Field(min_length=3, max_length=63, to_lower=True, regex="^[a-z][a-z0-9-]*[a-z0-9]$")
    zone: str
    pipeline_file: str
    pipeline_function: Optional[str]
    pipeline_params: Union[Path, Dict[str, Any], None]
    docker_image_ref: List[str] = []
    schedule: Optional[CloudSchedulerModel]
    tensorboard_ref: Optional[str]
    network: Optional[str]
    notification_channels_ref: List[str] = []
    sla_hours: Optional[float]



import pathlib
from pathlib import Path

import typer

from wanna.cli.plugins.base_plugin import BasePlugin
from wanna.cli.plugins.common_options import instance_name_option, profile_name_option, wanna_file_option
from wanna.core.services.managed_notebook import ManagedNotebookService
from wanna.core.utils.config_loader import load_config_from_yaml


class ManagedNotebookPlugin(BasePlugin):
    """
    Create, delete and more operations for managed Workbench (Jupyter notebook).
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_many(
            [
                self.delete,
                self.create,
                self.sync,
                self.report,
                self.build,
            ]
        )

    @staticmethod
    def delete(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("managed_notebook", "delete"),
    ) -> None:
        """
        Delete a Managed Workbench Notebook.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = ManagedNotebookService(config=config, workdir=workdir)
        nb_service.delete(instance_name)

    @staticmethod
    def create(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("managed_notebook", "create"),
    ) -> None:
        """
        Create a Managed Workbench Notebook.

        If there already is a notebook with the same name in the same location and project,
        you will be prompt if you want to delete the existing one and start a new one.

        When the notebook instance is created, you will be given a URL link to JupyterLab.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = ManagedNotebookService(config=config, workdir=workdir)
        nb_service.create(instance_name)

    @staticmethod
    def sync(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        force: bool = typer.Option(False, "--force", help="Synchronisation without prompt"),
    ) -> None:
        """
        Synchronize existing Managed Notebooks with wanna.yaml

        1. Reads current notebooks where label is defined per field wanna_project.name in wanna.yaml
        2. Does a diff between what is on GCP and what is on yaml
        3. Create the ones defined in yaml and missing in GCP
        4. Delete the ones in GCP that are not in wanna.yaml
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = ManagedNotebookService(config=config, workdir=workdir)
        nb_service.sync(force)

    @staticmethod
    def report(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
        instance_name: str = instance_name_option("managed_notebook", "report"),
    ) -> None:
        """
        Displays a link to the cost report per wanna_project and optionally per instance name.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = ManagedNotebookService(config=config, workdir=workdir)
        nb_service.report(
            instance_name=instance_name,
            wanna_project=config.wanna_project.name,
            wanna_resource="managed_notebook",
            gcp_project=config.gcp_profile.project_id,
            billing_id=config.wanna_project.billing_id,
            organization_id=config.wanna_project.organization_id,
        )

    @staticmethod
    def build(
        file: Path = wanna_file_option,
        profile_name: str = profile_name_option,
    ) -> None:
        """
        Validates build of notebooks as they are defined in wanna.yaml.
        """
        config = load_config_from_yaml(file, gcp_profile_name=profile_name)
        workdir = pathlib.Path(file).parent.resolve()
        nb_service = ManagedNotebookService(config=config, workdir=workdir)
        nb_service.build()



import logging
import sys
from typing import cast

import typer
from halo import Halo


class Spinner(Halo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_tty = sys.stdout.isatty()

    def __enter__(self):
        """Starts the spinner on a separate thread. For use in context managers.
        The spinner is actually started only in the interactive terminal (tty),
        if the environment is output only, we only print the start and end of the process.

        Returns
        -------
        self
        """
        return self.start() if self.is_tty else self.info()

    def __exit__(self, exception_type, exception_value, traceback):
        """Stops the spinner. For use in context managers."""
        if exception_value:
            self.text_color = typer.colors.RED
            self.fail()
        else:
            self.text_color = typer.colors.GREEN
            self.succeed()


class WannaLogger(logging.Logger):
    """
    This Logger supports all common logging library methods
    like .info, .warning or .debug.
    On top of that, we introduce new methods for visually
    more appealing printing to users.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set the logging config here if needed
        logging.basicConfig()

    @staticmethod
    def user_error(text, fg: str = typer.colors.RED, *args, **kwargs) -> None:
        typer.secho(f"✖ {text}", fg=fg, *args, **kwargs)

    @staticmethod
    def user_info(text, *args, **kwargs) -> None:
        typer.secho(f"ℹ {text}", *args, **kwargs)

    @staticmethod
    def user_success(text, fg: str = typer.colors.GREEN, *args, **kwargs) -> None:
        typer.secho(f"✔ {text}", fg=fg, *args, **kwargs)

    @staticmethod
    def user_spinner(text, *args, **kwargs) -> Spinner:
        return Spinner(text=text, *args, **kwargs)


def get_logger(name: str) -> WannaLogger:
    logging.setLoggerClass(WannaLogger)
    logger = cast(WannaLogger, logging.getLogger(name))
    return logger



from typing import Dict, Optional

from pydantic import BaseModel, Extra, root_validator, validator

from wanna.core.utils import validators
from wanna.core.utils.gcp import get_region_from_zone


class GCPProfileModel(BaseModel, extra=Extra.forbid):
    profile_name: str
    project_id: str
    zone: Optional[str]
    region: str
    labels: Optional[Dict[str, str]]
    bucket: str
    service_account: Optional[str]
    network: Optional[str]
    subnet: Optional[str]
    kms_key: Optional[str]
    docker_repository: str = "wanna"
    docker_registry: Optional[str]

    _ = validator("project_id", allow_reuse=True)(validators.validate_project_id)
    _ = validator("zone", allow_reuse=True)(validators.validate_zone)

    @root_validator(pre=True)
    def parse_region_from_zone(cls, values):  # pylint: disable=no-self-argument,no-self-use
        """
        In some cases, the zone is defined and region not.
        Region can be easily parsed from zone.
        """
        zone, region = (
            values.get("zone"),
            values.get("region"),
        )
        if (region is None) and (zone is not None):
            values["region"] = get_region_from_zone(zone)
        return values

    _ = validator("region", allow_reuse=True)(validators.validate_region)
